#pragma once

#include "GeometricObject.h"

class AxisAlignedBox : public GeometricObject
{
private:
	AxisAlignedBox()
	{
	}

	AxisAlignedBox(const AxisAlignedBox &b)
	{
		m_Pos[0] = b.m_Pos[0];
		m_Pos[1] = b.m_Pos[1];
	}

	AxisAlignedBox(const MyAlgebra::vec3 &pMin, const MyAlgebra::vec3 &pMax, Material *m)
	{
		m_Pos[0] = pMin;
		m_Pos[1] = pMax;
		material = m;
	}

public:
	static AxisAlignedBox *CreateGeometricObject()
	{
		AxisAlignedBox *obj = new AxisAlignedBox();
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static AxisAlignedBox *CreateGeometricObject(const AxisAlignedBox &b)
	{
		AxisAlignedBox *obj = new AxisAlignedBox(b);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static AxisAlignedBox *CreateGeometricObject(const MyAlgebra::vec3 &pMin, const MyAlgebra::vec3 &pMax, Material *m)
	{
		AxisAlignedBox *obj = new AxisAlignedBox(pMin, pMax, m);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	bool hit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const;
	bool shadowHit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const;

	void drawGL() const;	// for preview using OpenGL

	inline MyAlgebra::vec3 getMin() const { return m_Pos[0]; }
	inline MyAlgebra::vec3 getMax() const { return m_Pos[1]; }

	inline void setMin(const MyAlgebra::vec3 &p) { m_Pos[0] = p; }
	inline void setMax(const MyAlgebra::vec3 &p) { m_Pos[1] = p; }

private:
	MyAlgebra::vec3 m_Pos[2];

};

