#pragma once

// for experimental input/output of binary images in an original format

#include <iostream>
#include <fstream>
#include <string>

#include "ImageRect.h"
#include "data_path.h"

class BinaryImageIO
{
public:
	static void Init()
	{
		image_path.path.push_back("..");
		image_path.path.push_back("../..");
		image_path.path.push_back("../../..");
		image_path.path.push_back("images");
		image_path.path.push_back("../images");
		image_path.path.push_back("../../images");
	}

	static data_path image_path;

	template <class T>
	bool load(const char *filename, ImageRect<T> &img)
	{
		std::ifstream ifs(filename, std::ios::binary);

		if ( ! ifs)
		{
			std::cerr << __FUNCTION__ << ": cannot open file: " << filename << std::endl;
			return false;
		}

		std::string header;
		ifs >> header;

		if (header != "binary_image")
		{
			std::cerr << __FUNCTION__ << ": unknown file type: " << filename << std::endl;
			return false;
		}

		int w, h;
		ifs >> w >> h;

		if (w <= 0 || h <= 0)
		{
			std::cerr << __FUNCTION__ << ": invalid image size: " << w << "x" << h << std::endl;
			return false;
		}

		//std::cerr << __FUNCTION__ << ": image size: " << w << "x" << h << std::endl;

		std::string dataType;
		int nChannels;
		ifs >> dataType >> nChannels;

		//std::cerr << __FUNCTION__ << ": data type: " << dataType << ", " << nChannels << " channels" << std::endl;

		img.allocate(w,h);
		ifs.read((char *)img.getData(), sizeof(T)*w*h);

		std::cerr << __FUNCTION__ << ": file loaded: " << filename << " (" << w << "x" << h << ")" << std::endl;

		return true;
	}

	template <class T>
	bool save(const char *filename, const ImageRect<T> &img)
	{
		std::ofstream ofs(filename, std::ios::binary);

		if ( ! ofs)
		{
			std::cerr << __FUNCTION__ << ": cannot open file: " << filename << std::endl;
			return false;
		}

		const int w = img.getWidth();
		const int h = img.getHeight();

		ofs << "binary_image" << endl;
		ofs << w << " " << h << endl;
		ofs << getDataType(img) << " " << getNumChannels(img) << endl;
		ofs.write((const char *)img.getData(), sizeof(T)*w*h);

		std::cerr << __FUNCTION__ << ": file saved: " << filename << std::endl;

		return true;
	}

private:
	template <class T> const char *getDataType(const ImageRect<T> &img) const { return "uchar"; }
	template <> const char *getDataType(const ImageRect<UbyteLib::ubyte> &img) const { return "uchar"; }
	template <> const char *getDataType(const ImageRect<UbyteLib::ubyte3> &img) const { return "uchar"; }
	template <> const char *getDataType(const ImageRect<UbyteLib::ubyte4> &img) const { return "uchar"; }
	template <> const char *getDataType(const ImageRect<float> &img) const { return "float"; }
	template <> const char *getDataType(const ImageRect<MyAlgebra::vec3> &img) const { return "float"; }
	template <> const char *getDataType(const ImageRect<MyAlgebra::vec4> &img) const { return "float"; }

	template <class T> int getNumChannels(const ImageRect<T> &img) const { return 1; }
	template <> int getNumChannels(const ImageRect<UbyteLib::ubyte> &img) const { return 1; }
	template <> int getNumChannels(const ImageRect<UbyteLib::ubyte3> &img) const { return 3; }
	template <> int getNumChannels(const ImageRect<UbyteLib::ubyte4> &img) const { return 4; }
	template <> int getNumChannels(const ImageRect<float> &img) const { return 1; }
	template <> int getNumChannels(const ImageRect<MyAlgebra::vec3> &img) const { return 3; }
	template <> int getNumChannels(const ImageRect<MyAlgebra::vec4> &img) const { return 4; }
};

