#include "ConfigLoader.h"
#include "data_path.h"

#include "RecursiveRayTracer.h"

#include "Camera.h"
#include "Scene.h"

#include "Sphere.h"
#include "Plane.h"
#include "Triangle.h"
#include "TexturedPlane.h"
#include "TexturedSphere.h"
#include "SolidTexturedSphere.h"
#include "AxisAlignedBox.h"
#include "TriangleMesh.h"
#include "TriangleMeshUniformGrid.h"
#include "MetaballSurface.h"

#include "PointLight.h"
#include "DirectionalLight.h"
#include "EnvironmentLight.h"

#include "PseudoNormalColorMaterial.h"
#include "DiffuseMaterial.h"
#include "PhongMaterial.h"
#include "PerfectSpecularMaterial.h"
#include "SpecularRefractionMaterial.h"
#include "TexturedMaterial.h"

#include "CheckeredTexture.h"
#include "SolidCheckeredTexture.h"
#include "ImageTexture.h"
#include "NoiseTexture.h"
#include "MarbleTexture.h"
#include "WoodenTexture.h"
#include "TurbulenceTexture.h"
#include "DerivTurbulenceTexture.h"

#include "CheckeredTexture.h"
#include "SolidCheckeredTexture.h"
#include "NoiseTexture.h"
#include "MarbleTexture.h"
#include "WoodenTexture.h"
#include "TurbulenceTexture.h"
#include "DerivTurbulenceTexture.h"
#include "ImageTexture.h"

#include "BoxFilter.h"

#include "SingleSampler.h"
#include "RandomSampler.h"
#include "Regular2x2Sampler.h"
#include "Regular3x3Sampler.h"
#include "RegularNxNSampler.h"
#include "JitterSampler.h"
#include "MultiJitterSampler.h"

#include "EnvironmentMap.h"
#include "EnvironmentLightGenerator.h"

#include "MeshLoader.h"

#include "MiscSettings.h"

#include "CameraViewWindow.h"

using namespace std;
using namespace MyAlgebra;

extern RayTracer *g_pRayTracer;
extern Scene g_Scene;
extern Camera g_Camera;
extern ImageRect<vec3> g_ScreenBuffer;

extern vector<int> g_TracedGridCells;

extern MiscSettings g_MiscSettings;

extern CameraViewWindow g_CameraViewWindow;

map<string, GeometricObject *> ConfigLoader::s_GeometryDictionary;
map<string, Material *> ConfigLoader::s_MaterialDictionary;

bool ConfigLoader::load( const char *filename, bool isVerbose)
{
	m_IsVerbose = isVerbose;

	data_path::filepath_separator = '\\';

	s_GeometryDictionary.clear();
	s_MaterialDictionary.clear();

	m_Objects.clear();
	m_LightSources.clear();

	m_pEnvironmentMap = 0;

	m_MiscSettings = g_MiscSettings;

	m_ErrorFlag = false;

	if ( ! includeFile( filename ) )
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": cannot open file: " << filename << endl;

		return false;
	}

	m_MiscSettings.m_SceneFilename = filename;

	// ǂݍ񂾒lݒ 

	g_Scene.clearScene();

	if (m_pEnvironmentMap)
		g_Scene.setEnvironmentMap( m_pEnvironmentMap );

	for (int oi=0; oi<(int)m_Objects.size(); oi++)
		g_Scene.addObject( m_Objects[oi] );

	for (int li=0; li<(int)m_LightSources.size(); li++)
		g_Scene.addLight( m_LightSources[li] );

	g_MiscSettings = m_MiscSettings;

	//g_ScreenBuffer.allocate( m_MiscSettings.m_ScreenWidth, m_MiscSettings.m_ScreenHeight );
	glutSetWindow( g_CameraViewWindow.getWindowID() );
	glutReshapeWindow( m_MiscSettings.m_ScreenWidth, m_MiscSettings.m_ScreenHeight );

	//g_Camera.setLookAt( m_MiscSettings.m_CameraPosition, m_MiscSettings.m_CameraCenter, m_MiscSettings.m_CameraUpDir);
	g_Camera.setModelviewMatrix( m_MiscSettings.m_Modelview.mat_array );
	g_Camera.setFieldOfView( m_MiscSettings.m_CameraFovy );

	//// debug
	//{
	//	mat4 M;
	//	g_Camera.getModelviewMatrix( M.mat_array );

	//	cerr << __FUNCTION__ << ": input modelview: " << m_MiscSettings.m_Modelview << endl;
	//	cerr << __FUNCTION__ << ": output modelview: " << M << endl;
	//}
	//{
	//	vec3 pos, center, up;
	//	g_Camera.getLookAt(pos, center, up);

	//	cerr << __FUNCTION__ << ": camera input: eye=" << m_MiscSettings.m_CameraPosition << ", center="
	//		 << m_MiscSettings.m_CameraCenter << ", up=" << m_MiscSettings.m_CameraUpDir << endl;
	//	cerr << __FUNCTION__ << ": camera output: eye=" << pos << ", center="
	//		<< center << ", up=" << up << endl;
	//}

	if ( ! m_MiscSettings.m_pSampler)
	{
		cerr << __FUNCTION__ << ": setting single sampler" << endl;
		m_MiscSettings.m_pSampler = SingleSampler::Create();
	}

	g_pRayTracer->registerImageSampler( m_MiscSettings.m_pSampler );
	//g_pRayTracer->registerImageSampler( SingleSampler::Create() );
	m_MiscSettings.m_pSampler = 0;

	g_TracedGridCells.clear();

	cerr << __FUNCTION__ << ": \"" << filename << "\" loaded" << endl;

	return true;
}

// private functions --------------------------------------

bool ConfigLoader::includeFile( const char *filename )
{
	if (m_ErrorFlag) return false;

	data_path scenePath;
	scenePath.path.push_back(".");
	scenePath.path.push_back("..");
	scenePath.path.push_back("../..");
	scenePath.path.push_back("../../..");
	scenePath.path.push_back("Config");
	scenePath.path.push_back("../Config");
	scenePath.path.push_back("../../Config");
	scenePath.path.push_back("../../../Config");

	if (m_IsVerbose)
		cerr << __FUNCTION__ << ": loading file: \"" << filename << "\"" << endl;

	const string baseDir = data_path::getBasedir(filename);

	ifstream ifs( scenePath.get_file( filename ).c_str() );

	if ( ! ifs)
	{
		cerr << __FUNCTION__ << ": cannot open file: " << filename << endl;
		return false;
	}

	Tokenizer tokenizer( &ifs );

	string header;
	float version;

	m_ErrorFlag |= tokenizer.readString(header);
	m_ErrorFlag |= tokenizer.readFloat(version);

	if ( "ExperimentalRayTracerConfig" != header)
	{
		cerr << __FUNCTION__ << ": invalid header: " << header << endl;
		return false;
	}

	if (m_IsVerbose)
		cerr << __FUNCTION__ << ": version: " << version << endl;

	while ( true )
	{
		m_ErrorFlag |= tokenizer.skipWhiteSpaces();

		if (m_ErrorFlag)
		{
			cerr << __FUNCTION__ << ": " << filename << ": something wrong happened" << endl;
			break;
		}

		if ( ifs.eof() )
		{
			//cerr << __FUNCTION__ << ": " << filename << ": EOF" << endl;
			break;
		}

		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "#include" == token )
		{
			string newFilename;
			m_ErrorFlag |= tokenizer.readQuotedString( newFilename );

			if ( "" != baseDir)
				newFilename = baseDir + '/' + newFilename;

			m_ErrorFlag |= ! includeFile( scenePath.get_file( newFilename ).c_str() );
		}
		else if ( "#define" == token )
		{
			string name;
			m_ErrorFlag |= tokenizer.readString( name );

			string type;
			m_ErrorFlag |= tokenizer.readString( type );

			if ( IsGeometryToken(type) )
			{
				GeometricObject *obj;
				m_ErrorFlag |= parseGeometry( type, tokenizer, &obj );

				if ( ! m_ErrorFlag)
				{
					s_GeometryDictionary.insert( pair<string, GeometricObject*>(name, obj) );
					if (m_IsVerbose)
						cerr << __FUNCTION__ << ": registered geometry: " << name << endl;
				}
			}
			else if ( IsMaterialToken(type) )
			{
				Material *mat;
				m_ErrorFlag |= parseMaterial( type, tokenizer, &mat );

				if ( ! m_ErrorFlag)
				{
					s_MaterialDictionary.insert( pair<string, Material*>(name, mat) );
					if (m_IsVerbose)
						cerr << __FUNCTION__ << ": registered material: " << name << endl;
				}
			}
			else
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": unsupported type for #define: " << type << endl;
				m_ErrorFlag = true;
			}
		}
		else if ( "global_config" == token )
		{
			m_ErrorFlag |= parseGlobalConfig(tokenizer);
		}
		else if ( "pixel_filter" == token )
		{
			m_ErrorFlag |= parsePixelFilter(tokenizer);
		}
		else if ( "camera" == token )
		{
			m_ErrorFlag |= parseCamera(tokenizer);
		}
		else if ( "screen" == token )
		{
			m_ErrorFlag |= parseScreen(tokenizer);
		}
		else if ( "point_light" == token )
		{
			m_ErrorFlag |= parsePointLight(tokenizer);
		}
		else if ( "directional_light" == token )
		{
			m_ErrorFlag |= parseDirectionalLight(tokenizer);
		}
		else if ( "environment_light" == token )
		{
			m_ErrorFlag |= parseEnvironmentLight(tokenizer);
		}
		else if ( IsGeometryToken(token) )
		{
			GeometricObject *obj = 0;
			parseGeometry( token, tokenizer, &obj );
			if ( obj )
				m_Objects.push_back( obj );
		}
		else if ( IsTransformationToken(token) )
		{
			GeometryInstance *obj = GeometryInstance::CreateGeometricObject();
			parseTransformation( token, tokenizer, &obj );
			if ( ! m_ErrorFlag )
				m_Objects.push_back( obj );
			else
				delete obj;
		}
		else
		{
			cerr << __FUNCTION__ << ": unsupported token \"" << token << "\" at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": \"" << filename << "\" loaded" << endl;
		return true;
	}

	return false;
}

bool ConfigLoader::IsGeometryToken( const string &token )
{
	if ( "sphere" == token || "plane" == token || "box" == token || "cylinder" == token || "cone" == token ||
		"triangle" == token || "triangle_mesh" == token || "metaball_surface" == token )
		return true;

	map<string, GeometricObject *>::const_iterator itr = s_GeometryDictionary.find(token);

	return ( itr != s_GeometryDictionary.end() );
}

bool ConfigLoader::IsTransformationToken( const std::string &token )
{
	if ( "rotate" == token || "scale" == token || "translate" == token || "transform" == token )
		return true;

	return false;
}

bool ConfigLoader::IsMaterialToken( const string &token )
{
	if ( "pseudo_normal_color_material" == token || "ambient_material" == token || "diffuse_material" == token ||
		"phong_material" == token || "mirror_material" == token || "refract_material" == token ||
		"checker_texture_material" == token || "solid_checker_texture_material" == token ||
		"noise_texture_material" == token || "marble_texture_material" == token ||
		"wooden_texture_material" == token || "image_texture_material" == token ||
		"turbulence_texture_material" == token || "dturbulence_texture_material" == token )
		return true;

	map<string, Material *>::const_iterator itr = s_MaterialDictionary.find(token);

	return ( itr != s_MaterialDictionary.end() );
}

bool ConfigLoader::parseGlobalConfig( Tokenizer &tokenizer )
{
	int maxRecursion;

	m_ErrorFlag |= tokenizer.requireChar('{');	
	m_ErrorFlag |= tokenizer.requireString("max_recursion");
	m_ErrorFlag |= tokenizer.requireChar('=');
	m_ErrorFlag |= tokenizer.readInteger(maxRecursion);
	m_ErrorFlag |= tokenizer.requireChar(';');
	m_ErrorFlag |= tokenizer.requireChar('}');

	if ( ! m_ErrorFlag)
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": max_recursion = " << maxRecursion << endl;

		m_MiscSettings.m_RecursionDepth = maxRecursion;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parsePixelFilter( Tokenizer &tokenizer )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	string filterType = "";
	string samplerType = "";
	int nSamplesPerPixel = 1;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "filter_type" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readString(filterType);
			m_ErrorFlag |= tokenizer.requireChar(';');
		}
		else if ( "sampler_type" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readString(samplerType);
			m_ErrorFlag |= tokenizer.requireChar(';');
		}
		else if ( "num_samples_per_pixel" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readInteger(nSamplesPerPixel);
			m_ErrorFlag |= tokenizer.requireChar(';');
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		if ( "box" == filterType)
			m_MiscSettings.m_PixelFilterType = Filter::Box_Filter;

		if ("regular" == samplerType)
		{
			//if ( m_MiscSettings.m_pSampler ) delete m_MiscSettings.m_pSampler;
			m_MiscSettings.m_pSampler = RegularNxNSampler::Create();

			if ( m_IsVerbose )
				cerr << __FUNCTION__ << ": regular sampler" << endl;
		}
		else if ("random" == samplerType)
		{
			//if ( m_MiscSettings.m_pSampler ) delete m_MiscSettings.m_pSampler;
			m_MiscSettings.m_pSampler = RandomSampler::Create();

			if ( m_IsVerbose )
				cerr << __FUNCTION__ << ": random sampler" << endl;
		}
		else if ("jitter" == samplerType)
		{
			//if ( m_MiscSettings.m_pSampler ) delete m_MiscSettings.m_pSampler;
			m_MiscSettings.m_pSampler = JitterSampler::Create();

			if ( m_IsVerbose )
				cerr << __FUNCTION__ << ": jitter sampler" << endl;
		}
		else if ("multi_jitter" == samplerType)
		{
			//if ( m_MiscSettings.m_pSampler ) delete m_MiscSettings.m_pSampler;
			m_MiscSettings.m_pSampler = MultiJitterSampler::Create();

			if ( m_IsVerbose )
				cerr << __FUNCTION__ << ": multi-jitter sampler" << endl;
		}
		else if ("regular2x2" == samplerType)
		{
			if ( 4 != nSamplesPerPixel)
			{
				if ( m_IsVerbose )
					cerr << __FUNCTION__ << ": regular2x2 sampler, enforcing # of sample to be 4" << endl;
				nSamplesPerPixel = 4;
			}

			//if ( m_MiscSettings.m_pSampler ) delete m_MiscSettings.m_pSampler;
			m_MiscSettings.m_pSampler = Regular2x2Sampler::Create();

			if ( m_IsVerbose )
				cerr << __FUNCTION__ << ": regular2x2 sampler" << endl;
		}
		else if ("regular3x3" == samplerType)
		{
			if ( 9 != nSamplesPerPixel)
			{
				if ( m_IsVerbose )
					cerr << __FUNCTION__ << ": regular2x2 sampler, enforcing # of sample to be 9" << endl;
				nSamplesPerPixel = 9;
			}

			//if ( m_MiscSettings.m_pSampler ) delete m_MiscSettings.m_pSampler;
			m_MiscSettings.m_pSampler = Regular3x3Sampler::Create();

			if ( m_IsVerbose )
				cerr << __FUNCTION__ << ": regular3x3 sampler" << endl;
		}
		else // if ("single" == samplerType)
		{
			if ( 1 != nSamplesPerPixel)
			{
				if ( m_IsVerbose )
					cerr << __FUNCTION__ << ": single sampler, enforcing # of sample to be 1" << endl;
				nSamplesPerPixel = 1;
			}

			//if ( m_MiscSettings.m_pSampler ) delete m_MiscSettings.m_pSampler;
			m_MiscSettings.m_pSampler = SingleSampler::Create();

			if ( m_IsVerbose )
				cerr << __FUNCTION__ << ": single sampler" << endl;
		}

		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": num_samples_per_pixel = " << nSamplesPerPixel << endl;

		RecursiveRayTracer::s_NumSamplesPerPixel = nSamplesPerPixel;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseCamera( Tokenizer &tokenizer )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	bool isModelviewSpecified = false;
	vec3 position, center, updir;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "position" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(position);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": position = " << position << endl;
			}
		}
		else if ( "center" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(center);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": center = " << center << endl;
			}
		}
		else if ( "updir" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(updir);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": updir = " << updir << endl;
			}
		}
		else if ( "modelview" == token )
		{
			vec4 col0, col1, col2, col3;

			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec4( col0 );
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readVec4( col1 );
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readVec4( col2 );
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readVec4( col3 );
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				isModelviewSpecified = true;
				m_MiscSettings.m_Modelview.set_col(0, col0);
				m_MiscSettings.m_Modelview.set_col(1, col1);
				m_MiscSettings.m_Modelview.set_col(2, col2);
				m_MiscSettings.m_Modelview.set_col(3, col3);

				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": modelview = " << m_MiscSettings.m_Modelview << endl;
			}
		}
		else if ( "fovy" == token )
		{
			float fovy;

			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(fovy);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": fovy = " << fovy << endl;

				m_MiscSettings.m_CameraFovy = fovy;
			}
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag && ! isModelviewSpecified)
	{
		Camera tmpCamera;
		tmpCamera.setLookAt(position, center, updir);
		tmpCamera.getModelviewMatrix( m_MiscSettings.m_Modelview.mat_array );

		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": modelview = " << m_MiscSettings.m_Modelview << endl;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseScreen( Tokenizer &tokenizer )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "width" == token )
		{
			int width;

			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readInteger(width);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": width = " << width << endl;

				m_MiscSettings.m_ScreenWidth = width;
			}
		}
		else if ( "height" == token )
		{
			int height;

			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readInteger(height);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": height = " << height << endl;

				m_MiscSettings.m_ScreenHeight = height;
			}
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parsePointLight( Tokenizer &tokenizer )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	vec3 position, color;
	bool isPositionSet = false;
	bool isColorSet = false;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "position" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(position);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": position = " << position << endl;

			if ( ! m_ErrorFlag)
				isPositionSet = true;
		}
		else if ( "color" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color = " << color << endl;

			if ( ! m_ErrorFlag)
				isColorSet = true;
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( isPositionSet && isColorSet )
		m_LightSources.push_back( new PointLight(position, color) );
	else
	{
		cerr << __FUNCTION__ << ": error: both \"position\" and \"color\" must be specified" << endl;
		m_ErrorFlag = true;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseDirectionalLight( Tokenizer &tokenizer )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	vec3 direction, color;
	bool isDirectionSet = false;
	bool isColorSet = false;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "direction" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(direction);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				direction.normalize();

				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": direction = " << direction << endl;

				isDirectionSet = true;
			}
		}
		else if ( "color" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": color = " << color << endl;

				isColorSet = true;
			}
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( isDirectionSet && isColorSet )
		m_LightSources.push_back( new DirectionalLight(direction, color) );
	else
	{
		cerr << __FUNCTION__ << ": error: both \"direction\" and \"color\" must be specified" << endl;
		m_ErrorFlag = true;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseEnvironmentLight( Tokenizer &tokenizer )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	string filename;
	float hdrGamma = m_MiscSettings.m_HdrGamma;
	int nIblSamples = m_MiscSettings.m_NumIblSamples;
	float lightScale = m_MiscSettings.m_IblLightScale;
	bool isFilenameSet = false;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "filename" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readQuotedString(filename);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": filename = " << filename << endl;

				isFilenameSet = true;
			}
		}
		else if ( "hdr_gamma" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(hdrGamma);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": hdrGamma = " << hdrGamma << endl;
		}
		else if ( "light_scale" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(lightScale);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": lightScale = " << lightScale << endl;
		}
		else if ( "num_ibl_samples" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readInteger(nIblSamples);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": nIblSamples = " << nIblSamples << endl;
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag )
	{
		if (isFilenameSet)
		{
			m_MiscSettings.m_EnvironmentMapFilename = filename;
			m_MiscSettings.m_HdrGamma = hdrGamma;
			m_MiscSettings.m_NumIblSamples = nIblSamples;
			m_MiscSettings.m_IblLightScale = lightScale;

			m_pEnvironmentMap = new EnvironmentMap();

			m_pEnvironmentMap->setHDRGamma(hdrGamma);

			if ( m_pEnvironmentMap->load(filename.c_str()) && nIblSamples )
			{
				EnvironmentLightGenerator generator;
				generator.generate(nIblSamples, m_pEnvironmentMap, m_LightSources);

				EnvironmentLight::s_LuminanceScaleCoefficient = lightScale;
			}
			else
			{
				delete m_pEnvironmentMap;
				m_pEnvironmentMap = 0;
			}
		}
		else
		{
			cerr << __FUNCTION__ << ": error: \"filename\" must be specified" << endl;
			m_ErrorFlag = true;
		}
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseGeometry( const string &token, Tokenizer &tokenizer, GeometricObject **obj )
{
	if ( "sphere" == token ) return parseSphere(tokenizer, obj);
	else if ( "plane" == token ) return parsePlane(tokenizer, obj);
	else if ( "box" == token ) return parseBox(tokenizer, obj);
	else if ( "cylinder" == token ) return parseCylinder(tokenizer, obj);
	else if ( "cone" == token ) return parseCone(tokenizer, obj);
	else if ( "triangle" == token ) return parseTriangle(tokenizer, obj);
	else if ( "triangle_mesh" == token ) return parseTriangleMesh(tokenizer, obj);
	else if ( "metaball_surface" == token ) return parseMetaballSurface(tokenizer, obj);

	if ( s_GeometryDictionary.find(token) != s_GeometryDictionary.end() )
	{
		map<string, GeometricObject *>::const_iterator itr = s_GeometryDictionary.find(token);
		*obj = itr->second;

		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": assign registered geometry: " << token << endl;

		return false;
	}

	return true;	// G[ 
}

bool ConfigLoader::parseSphere( Tokenizer &tokenizer, GeometricObject **obj )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	Material *mat = 0;
	vec3 center;
	float radius;
	bool isCenterSet = false;
	bool isRadiusSet = false;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "center" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(center);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": center = " << center << endl;

				isCenterSet = true;
			}
		}
		else if ( "radius" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(radius);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": radius = " << radius << endl;

				isRadiusSet = true;
			}
		}
		else if ( IsMaterialToken(token) )
		{
			m_ErrorFlag |= parseMaterial(token, tokenizer, &mat);
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
		if ( isCenterSet && isRadiusSet )
		{
			if ( ! mat)
			{
				cerr << __FUNCTION__ << ": material not set, default material used" << endl;
				*obj = Sphere::CreateGeometricObject(center, radius, m_MiscSettings.m_pDefaultMaterial );
			}
			else
			{
				if (mat->getMaterialType() == Material::Textured_Type)
				{
					const Texture *tex = ((TexturedMaterial *)mat)->getTexture();
					if ( tex->isSolidTexture() )
					{
						if (m_IsVerbose)
							cerr << __FUNCTION__ << ": solid textured sphere required" << endl;

						*obj = SolidTexturedSphere::CreateGeometricObject(center, radius, mat);
					}
					else
					{
						if (m_IsVerbose)
							cerr << __FUNCTION__ << ": textured sphere required" << endl;

						*obj = TexturedSphere::CreateGeometricObject(center, radius, mat);
					}
				}
				else
				{
					if (m_IsVerbose)
						cerr << __FUNCTION__ << ": ordinary sphere required" << endl;

					*obj = Sphere::CreateGeometricObject(center, radius, mat);
				}
			}
		}
		else
		{
			cerr << __FUNCTION__ << ": error: both \"center\" and \"radius\" must be specified" << endl;
			m_ErrorFlag = true;
		}

		return m_ErrorFlag;
}

bool ConfigLoader::parsePlane( Tokenizer &tokenizer, GeometricObject **obj )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	Material *mat = 0;
	vec3 normal;
	float texScale = 1.f;
	float distance;
	bool isNormalSet = false;
	bool isDistanceSet = false;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "normal" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(normal);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				normal.normalize();

				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": normal = " << normal << endl;

				isNormalSet = true;
			}
		}
		else if ( "tex_scale" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(texScale);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": tex_scale = " << texScale << endl;
			}
		}
		else if ( "distance" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(distance);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": distance = " << distance << endl;

				isDistanceSet = true;
			}
		}
		else if ( IsMaterialToken(token) )
		{
			m_ErrorFlag |= parseMaterial(token, tokenizer, &mat);
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
		if ( isNormalSet && isDistanceSet )
			if ( ! mat)
			{
				cerr << __FUNCTION__ << ": material not set, default material used" << endl;
				*obj = Plane::CreateGeometricObject(normal, distance, m_MiscSettings.m_pDefaultMaterial);
			}
			else
			{
				if (mat->getMaterialType() == Material::Textured_Type)
				{
					if (m_IsVerbose)
						cerr << __FUNCTION__ << ": textured plane required" << endl;

					*obj = TexturedPlane::CreateGeometricObject(normal, distance, mat);
					((TexturedPlane *)(*obj))->setTextureScale( texScale );
				}
				else
				{
					if (m_IsVerbose)
						cerr << __FUNCTION__ << ": ordinary plane required" << endl;

					*obj = Plane::CreateGeometricObject(normal, distance, mat);
				}
			}
		else
		{
			cerr << __FUNCTION__ << ": error: both \"normal\" and \"distance\" must be specified" << endl;
			m_ErrorFlag = true;
		}

		return m_ErrorFlag;
}

bool ConfigLoader::parseBox( Tokenizer &tokenizer, GeometricObject **obj )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	Material *mat = 0;
	vec3 min_position, max_position;
	bool isMinPositionSet = false;
	bool isMaxPositionSet = false;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "min_position" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(min_position);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": min_position = " << min_position << endl;

				isMinPositionSet = true;
			}
		}
		else if ( "max_position" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(max_position);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": max_position = " << max_position << endl;

				isMaxPositionSet = true;
			}
		}
		else if ( IsMaterialToken(token) )
		{
			m_ErrorFlag |= parseMaterial(token, tokenizer, &mat);
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
		if ( isMinPositionSet && isMaxPositionSet )
		{
			if ( ! mat)
			{
				cerr << __FUNCTION__ << ": material not set, default material used" << endl;
				mat = m_MiscSettings.m_pDefaultMaterial;
			}

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": textured box required" << endl;

			*obj = AxisAlignedBox::CreateGeometricObject(min_position, max_position, mat);
		}
		else
		{
			cerr << __FUNCTION__ << ": error: both \"min_position\" and \"max_position\" must be specified" << endl;
			m_ErrorFlag = true;
		}

		return m_ErrorFlag;
}

bool ConfigLoader::parseCylinder( Tokenizer &tokenizer, GeometricObject **obj )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	cerr << __FUNCTION__ << ": not supported yet" << endl;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "}" == token )
		{
			break;
		}
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseCone( Tokenizer &tokenizer, GeometricObject **obj )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	cerr << __FUNCTION__ << ": not supported yet" << endl;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "}" == token )
		{
			break;
		}
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseTriangle( Tokenizer &tokenizer, GeometricObject **obj )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	Material *mat = 0;
	vec3 vertices[3], normals[3];
	vec2 texCoords[3];

	bool isVerticesSet = false;
	bool isNormalsSpecified = false;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "vertices" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.requireChar('{');
			m_ErrorFlag |= tokenizer.readVec3(vertices[0]);
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readVec3(vertices[1]);
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readVec3(vertices[2]);
			m_ErrorFlag |= tokenizer.requireChar('}');
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag)
			{
				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": vertices = { " << vertices[0] << ", " << vertices[1] << ", " << vertices[2] << " }" << endl;

				isVerticesSet = true;
			}
		}
		else if ( "normals" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.requireChar('{');
			m_ErrorFlag |= tokenizer.readVec3(normals[0]);
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readVec3(normals[1]);
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readVec3(normals[2]);
			m_ErrorFlag |= tokenizer.requireChar('}');
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag )
			{
				isNormalsSpecified = true;

				normals[0].normalize();
				normals[1].normalize();
				normals[2].normalize();

				if (m_IsVerbose)
					cerr << __FUNCTION__ << ": normals = { " << normals[0] << ", " << normals[1] << ", " << normals[2] << " }" << endl;
			}
		}
		else if ( "tex_coords" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.requireChar('{');
			m_ErrorFlag |= tokenizer.readVec2(texCoords[0]);
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readVec2(texCoords[1]);
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readVec2(texCoords[2]);
			m_ErrorFlag |= tokenizer.requireChar('}');
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": texCoords = { " << texCoords[0] << ", " << texCoords[1] << ", " << texCoords[2] << " }" << endl;
		}
		else if ( IsMaterialToken(token) )
		{
			m_ErrorFlag |= parseMaterial(token, tokenizer, &mat);
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag )
	{
		if ( ! isNormalsSpecified)
		{
			const vec3 normal = normalize( (vertices[1]-vertices[0]) ^ (vertices[2]-vertices[0]) );
			normals[0] = normals[1] = normals[2] = normal;

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": note: normals computed: " << normal << endl;
		}

		//  vertices ZbgĂȂ΃G[ɂׂ
		// normals ͗^ĂȂ΂̏ŌvZׂ 

		if ( ! mat)
		{
			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": material not set, default material used" << endl;

			mat = m_MiscSettings.m_pDefaultMaterial;
		}

		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": triangle added" << endl;

		Triangle *newObj = Triangle::CreateGeometricObject();
		newObj->setVertex0( vertices[0] );
		newObj->setVertex1( vertices[1] );
		newObj->setVertex2( vertices[2] );
		newObj->setNormal0( normals[0] );
		newObj->setNormal1( normals[1] );
		newObj->setNormal2( normals[2] );
		newObj->setTexcoord0( texCoords[0] );
		newObj->setTexcoord1( texCoords[1] );
		newObj->setTexcoord2( texCoords[2] );
		newObj->setMaterial( mat );
		*obj = newObj;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseTriangleMesh( Tokenizer &tokenizer, GeometricObject **obj )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	Material *mat = 0;
	string filename = "";
	string shadingType;
	bool gridAccelerationEnabled = false;

	int gridXRes = 32;
	int gridYRes = 32;
	int gridZRes = 32;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "filename" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readQuotedString(filename);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": filename = " << filename << endl;
		}
		else if ( "shading_type" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readString(shadingType);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shadingType = " << shadingType << endl;
		}
		else if ( "grid_acceleration" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readBoolean(gridAccelerationEnabled);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": grid_acceleration = " << gridAccelerationEnabled << endl;
		}
		else if ( "grid_resolution" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.requireChar('(');
			m_ErrorFlag |= tokenizer.readInteger( gridXRes );
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readInteger( gridYRes );
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readInteger( gridZRes );
			m_ErrorFlag |= tokenizer.requireChar(')');
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": grid_resolution = (" << gridXRes << "," << gridYRes << "," << gridZRes << ")" << endl;
		}
		else if ( IsMaterialToken(token) )
		{
			m_ErrorFlag |= parseMaterial(token, tokenizer, &mat);
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! mat)
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": material not set, default material used" << endl;
		mat = m_MiscSettings.m_pDefaultMaterial;
	}

	if (m_IsVerbose)
		cerr << __FUNCTION__ << ": ordinary triangle mesh added" << endl;

	if (gridAccelerationEnabled)
	{
		TriangleMeshUniformGrid *newObj = TriangleMeshUniformGrid::CreateGeometricObject();

		if ("smooth" == shadingType)
			newObj->setShadingType( TriangleMeshUniformGrid::Smooth_Shading );
		else
			newObj->setShadingType( TriangleMeshUniformGrid::Flat_Shading );

		MeshLoader loader;
		m_ErrorFlag |= ! loader.load( filename.c_str(), *(TriangleMesh *)newObj );

		if ( ! m_ErrorFlag)
		{
			newObj->buidGrid(gridXRes, gridYRes, gridZRes);
			newObj->setMaterial( mat );
			*obj = newObj;
		}
	}
	else
	{
		TriangleMesh *newObj = TriangleMesh::CreateGeometricObject();

		if ("smooth" == shadingType)
			newObj->setShadingType( TriangleMesh::Smooth_Shading );
		else
			newObj->setShadingType( TriangleMesh::Flat_Shading );

		MeshLoader loader;
		m_ErrorFlag |= ! loader.load( filename.c_str(), *newObj );

		if ( ! m_ErrorFlag)
		{
			newObj->setMaterial( mat );
			*obj = newObj;
		}
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseMetaballSurface( Tokenizer &tokenizer, GeometricObject **obj )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	Material *mat = 0;
	MetaballSurface *pMetaballs = 0;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "metaball" == token )
		{
			vec3 center;
			float radius;

			m_ErrorFlag |= tokenizer.requireChar('{');	
			m_ErrorFlag |= tokenizer.readVec3(center);
			m_ErrorFlag |= tokenizer.requireChar(',');
			m_ErrorFlag |= tokenizer.readFloat(radius);	
			m_ErrorFlag |= tokenizer.requireChar('}');	
			m_ErrorFlag |= tokenizer.requireChar(';');

			if ( ! m_ErrorFlag )
			{
				if ( 0 == pMetaballs) pMetaballs = MetaballSurface::CreateGeometricObject();

				Metaball m(center, radius);
				pMetaballs->addMetaball( m );
			}
		}
		else if ( IsMaterialToken(token) )
		{
			m_ErrorFlag |= parseMaterial(token, tokenizer, &mat);
		}
		else if ( "}" == token )
		{
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! mat)
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": material not set, default material used" << endl;
		mat = m_MiscSettings.m_pDefaultMaterial;
	}

	if ( pMetaballs )
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": " << pMetaballs->getNumMetaballs() << " metaballs added" << endl;

		pMetaballs->setMaterial( mat );

		if ( ! m_ErrorFlag)
			*obj = pMetaballs;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseTransformation( const string &token, Tokenizer &tokenizer, GeometryInstance **obj )
{
	if ( "transform" == token ) return parseTransform( tokenizer, obj );
	else if ( "scale" == token ) return parseScale( tokenizer, obj );
	else if ( "rotate" == token ) return parseRotate( tokenizer, obj );
	else if ( "translate" == token ) return parseTranslate( tokenizer, obj );

	if ( IsGeometryToken( token ) )
	{
		GeometricObject *newObj = 0;
		m_ErrorFlag |= parseGeometry( token, tokenizer, &newObj );

		if ( newObj )
			(*obj)->setObject( newObj );

		return m_ErrorFlag;
	}

	return true;
}

bool ConfigLoader::parseTransform( Tokenizer &tokenizer, GeometryInstance **obj )
{
	vec4 columns[4];

	m_ErrorFlag |= tokenizer.requireChar('(');
	m_ErrorFlag |= tokenizer.readVec4( columns[0] );
	m_ErrorFlag |= tokenizer.requireChar(',');
	m_ErrorFlag |= tokenizer.readVec4( columns[1] );
	m_ErrorFlag |= tokenizer.requireChar(',');
	m_ErrorFlag |= tokenizer.readVec4( columns[2] );
	m_ErrorFlag |= tokenizer.requireChar(',');
	m_ErrorFlag |= tokenizer.readVec4( columns[3] );
	m_ErrorFlag |= tokenizer.requireChar(',');

	string token;
	m_ErrorFlag |= tokenizer.readString(token);

	m_ErrorFlag |= parseTransformation( token, tokenizer, obj );
	m_ErrorFlag |= tokenizer.requireChar(')');

	if ( ! m_ErrorFlag)
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": matrix = [" << columns[0] << ", " << columns[1] << ", "
			<< columns[2] << ", " << columns[3] << "]" << endl; 

		(*obj)->transform( columns[0], columns[1], columns[2], columns[3] );
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseScale( Tokenizer &tokenizer, GeometryInstance **obj )
{
	vec3 scale;

	m_ErrorFlag |= tokenizer.requireChar('(');
	m_ErrorFlag |= tokenizer.readFloats(scale.vec_array, 3);
	m_ErrorFlag |= tokenizer.requireChar(',');

	string token;
	m_ErrorFlag |= tokenizer.readString(token);
	m_ErrorFlag |= parseTransformation( token, tokenizer, obj );
	m_ErrorFlag |= tokenizer.requireChar(')');

	if ( ! m_ErrorFlag)
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": scale = " << scale << endl; 

		(*obj)->scale( scale.x, scale.y, scale.z );
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseRotate( Tokenizer &tokenizer, GeometryInstance **obj )
{
	vec4 angleAxes;

	m_ErrorFlag |= tokenizer.requireChar('(');
	m_ErrorFlag |= tokenizer.readFloats(angleAxes.vec_array, 4);
	m_ErrorFlag |= tokenizer.requireChar(',');

	string token;
	m_ErrorFlag |= tokenizer.readString(token);
	m_ErrorFlag |= parseTransformation( token, tokenizer, obj );
	m_ErrorFlag |= tokenizer.requireChar(')');

	if ( ! m_ErrorFlag)
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": angleAxes = " << angleAxes << endl; 

		(*obj)->rotate( angleAxes.x, angleAxes.y, angleAxes.z, angleAxes.w );
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseTranslate( Tokenizer &tokenizer, GeometryInstance **obj )
{
	vec3 translate;

	m_ErrorFlag |= tokenizer.requireChar('(');
	m_ErrorFlag |= tokenizer.readFloats(translate.vec_array, 3);
	m_ErrorFlag |= tokenizer.requireChar(',');

	string token;
	m_ErrorFlag |= tokenizer.readString(token);
	m_ErrorFlag |= parseTransformation( token, tokenizer, obj );
	m_ErrorFlag |= tokenizer.requireChar(')');

	if ( ! m_ErrorFlag)
	{
		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": translate = " << translate << endl; 

		(*obj)->translate( translate );
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseMaterial( const string &token, Tokenizer &tokenizer, Material **mat )
{
	if ( "pseudo_normal_color_material" == token ) return parsePseudoNormalColor( tokenizer, mat );
	else if ( "ambient_material" == token ) return parseAmbient( tokenizer, mat );
	else if ( "diffuse_material" == token ) return parseDiffuse( tokenizer, mat );
	else if ( "phong_material" == token ) return parsePhong( tokenizer, mat );
	else if ( "mirror_material" == token ) return parseMirror( tokenizer, mat );
	else if ( "refract_material" == token ) return parseRefract( tokenizer, mat );
	else if ( "checker_texture_material" == token ) return parseCheckerTexture( tokenizer, mat );
	else if ( "solid_checker_texture_material" == token ) return parseSolidCheckerTexture( tokenizer, mat );
	else if ( "noise_texture_material" == token ) return parseNoiseTexture( tokenizer, mat );
	else if ( "marble_texture_material" == token ) return parseMarbleTexture( tokenizer, mat );
	else if ( "wooden_texture_material" == token ) return parseWoodenTexture( tokenizer, mat );
	else if ( "image_texture_material" == token ) return parseImageTexture( tokenizer, mat );
	else if ( "turbulence_texture_material" == token ) return parseTurbulenceTexture( tokenizer, mat );
	else if ( "dturbulence_texture_material" == token) return parseDturbulenceTexture( tokenizer, mat );

	map<string, Material *>::iterator itr = s_MaterialDictionary.find(token);
	if ( itr != s_MaterialDictionary.end() )
	{
		*mat = itr->second;

		if (m_IsVerbose)
			cerr << __FUNCTION__ << ": assign registered material: " << token << endl;

		m_ErrorFlag |= tokenizer.requireChar(';');

		return m_ErrorFlag;
	}

	cerr << __FUNCTION__ << ": unknown material type: " << token << endl;

	return true;
}

bool ConfigLoader::parsePseudoNormalColor( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar(';');

	*mat = PseudoNormalColorMaterial::CreateMaterial();

	return m_ErrorFlag;
}

bool ConfigLoader::parseAmbient( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');	

	vec3 ambient(0,0,0);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		*mat = Material::CreateMaterial();
		(*mat)->setAmbientCoeff( ambient );
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseDiffuse( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), diffuse(0,0,0);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "diffuse" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(diffuse);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": diffuse = " << diffuse << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		DiffuseMaterial *newMat = DiffuseMaterial::CreateMaterial();
		newMat->setAmbientCoeff( ambient );
		newMat->setDiffuseCoeff( diffuse );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parsePhong( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), diffuse(0,0,0), specular(0,0,0);
	float shininess = 32.f;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "diffuse" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(diffuse);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": diffuse = " << diffuse << endl;
		}
		else if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "shininess" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(shininess);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shininess = " << shininess << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		PhongMaterial *newMat = PhongMaterial::CreateMaterial();
		newMat->setAmbientCoeff( ambient );
		newMat->setDiffuseCoeff( diffuse );
		newMat->setPhongCoeff( specular );
		newMat->setShininess( shininess );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseMirror( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 specular(0,0,0);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		PerfectSpecularMaterial *newMat = PerfectSpecularMaterial::CreateMaterial();
		newMat->setSpecularCoeff( specular );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseRefract( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 specular(0,0,0);
	float refractionIndex = 1.33f;

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "index" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(refractionIndex);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": refractionIndex = " << refractionIndex << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		SpecularRefractionMaterial *newMat = SpecularRefractionMaterial::CreateMaterial();
		newMat->setRefractionIndex( refractionIndex );
		newMat->setSpecularCoeff( specular );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseCheckerTexture( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), specular(0,0,0);
	float shininess = 32.f;

	int nURepeat = 5;
	int nVRepeat = 5;
	vec3 color0(0,0,0), color1(1,1,1);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "num_u_repeat" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readInteger(nURepeat);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": nURepeat = " << nURepeat << endl;
		}
		else if ( "num_v_repeat" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readInteger(nVRepeat);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": nVRepeat = " << nVRepeat << endl;
		}
		else if ( "color0" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color0);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color0 << endl;
		}
		else if ( "color1" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color1);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color1 << endl;
		}
		else if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "shininess" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(shininess);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shininess = " << shininess << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		CheckeredTexture *tex = CheckeredTexture::CreateTexture();
		tex->setUNumRepeats( nURepeat );
		tex->setVNumRepeats( nVRepeat );
		tex->setColors( color0, color1 );

		TexturedMaterial *newMat = TexturedMaterial::CreateMaterial();
		newMat->setAmbientCoeff( ambient );
		newMat->setPhongCoeff( specular );
		newMat->setShininess( shininess );
		newMat->setTexture( tex );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseSolidCheckerTexture( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), specular(0,0,0);
	float shininess = 32.f;

	int nURepeat = 5;
	int nVRepeat = 5;
	int nWRepeat = 5;
	vec3 color0(0,0,0), color1(1,1,1);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "num_u_repeat" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readInteger(nURepeat);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": num_u_repeat = " << nURepeat << endl;
		}
		else if ( "num_v_repeat" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readInteger(nVRepeat);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": num_v_repeat = " << nVRepeat << endl;
		}
		else if ( "num_w_repeat" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readInteger(nWRepeat);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": num_w_repeat = " << nWRepeat << endl;
		}
		else if ( "color0" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color0);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color0 << endl;
		}
		else if ( "color1" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color1);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color1 << endl;
		}
		else if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "shininess" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(shininess);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shininess = " << shininess << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		SolidCheckeredTexture *tex = SolidCheckeredTexture::CreateTexture();
		tex->setUNumRepeats( nURepeat );
		tex->setVNumRepeats( nVRepeat );
		tex->setWNumRepeats( nWRepeat );
		tex->setColors( color0, color1 );

		TexturedMaterial *newMat = TexturedMaterial::CreateMaterial();
		newMat->setAmbientCoeff( ambient );
		newMat->setPhongCoeff( specular );
		newMat->setShininess( shininess );
		newMat->setTexture( tex );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseNoiseTexture( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), specular(0,0,0);
	float shininess = 32.f;

	float scale = 10.f;
	vec3 color0(0,0,0), color1(1,1,1);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "scale" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(scale);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": scale = " << scale << endl;
		}
		else if ( "color0" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color0);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color0 << endl;
		}
		else if ( "color1" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color1);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color1 << endl;
		}
		else if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "shininess" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(shininess);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shininess = " << shininess << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		NoiseTexture *tex = NoiseTexture::CreateTexture();
		tex->setScale( scale );
		tex->setColor0( color0 );
		tex->setColor1( color1 );

		TexturedMaterial *newMat = TexturedMaterial::CreateMaterial();
		newMat->setAmbientCoeff( ambient );
		newMat->setPhongCoeff( specular );
		newMat->setShininess( shininess );
		newMat->setTexture( tex );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseMarbleTexture( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), specular(0,0,0);
	float shininess = 32.f;

	float noiseScale = 20.f;
	float noiseStripe = 0.15f;
	vec3 color0(0.8,0.8,0.8), color1(0.4,0.2,0.1), color2(0.06,0.04,0.02);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "noise_scale" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(noiseScale);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": noise_scale = " << noiseScale << endl;
		}
		else if ( "noise_stripe" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(noiseStripe);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": noise_stripe = " << noiseStripe << endl;
		}
		else if ( "color0" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color0);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color0 << endl;
		}
		else if ( "color1" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color1);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color1 = " << color1 << endl;
		}
		else if ( "color2" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color2);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color2 = " << color2 << endl;
		}
		else if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "shininess" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(shininess);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shininess = " << shininess << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		MarbleTexture *tex = MarbleTexture::CreateTexture();
		tex->setScale( noiseScale );
		tex->setStripesPerUnit( noiseStripe );
		tex->setColor0( color0 );
		tex->setColor1( color1 );
		tex->setColor1( color2 );

		TexturedMaterial *newMat = TexturedMaterial::CreateMaterial();
		newMat->setAmbientCoeff( ambient );
		newMat->setPhongCoeff( specular );
		newMat->setShininess( shininess );
		newMat->setTexture( tex );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseWoodenTexture( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), specular(0,0,0);
	float shininess = 32.f;

	float noiseScale = 20.f;
	float noiseStripe = 0.15f;
	vec3 color0(0.2, 0.1, 0.1), color1(0.8, 0.6, 0.4);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "noise_scale" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(noiseScale);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": noise_scale = " << noiseScale << endl;
		}
		else if ( "noise_stripe" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(noiseStripe);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": noise_stripe = " << noiseStripe << endl;
		}
		else if ( "color0" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color0);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color0 << endl;
		}
		else if ( "color1" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color1);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color1 = " << color1 << endl;
		}
		else if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "shininess" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(shininess);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shininess = " << shininess << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		WoodenTexture *tex = WoodenTexture::CreateTexture();
		tex->setScale( noiseScale );
		tex->setStripesPerUnit( noiseStripe );
		tex->setColor0( color0 );
		tex->setColor1( color1 );

		TexturedMaterial *newMat = TexturedMaterial::CreateMaterial();
		newMat->setAmbientCoeff( ambient );
		newMat->setPhongCoeff( specular );
		newMat->setShininess( shininess );
		newMat->setTexture( tex );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseImageTexture( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), specular(0,0,0);
	float shininess = 32.f;

	string filename = "";

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "filename" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readQuotedString(filename);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": filename = " << filename << endl;
		}
		else if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "shininess" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(shininess);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shininess = " << shininess << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		ImageTexture *tex = ImageTexture::CreateTexture();
		m_ErrorFlag |= ! tex->loadImage(filename);

		if ( ! m_ErrorFlag)
		{
			TexturedMaterial *newMat = TexturedMaterial::CreateMaterial();
			newMat->setAmbientCoeff( ambient );
			newMat->setPhongCoeff( specular );
			newMat->setShininess( shininess );
			newMat->setTexture( tex );
			*mat = newMat;
		}
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseTurbulenceTexture( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), specular(0,0,0);
	float shininess = 32.f;

	float noiseScale = 3.f;
	vec3 color0(0.8, 0.8, 0.8), color1(0.0, 0.0, 1.0);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "noise_scale" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(noiseScale);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": noise_scale = " << noiseScale << endl;
		}
		else if ( "color0" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color0);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color0 << endl;
		}
		else if ( "color1" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color1);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color1 = " << color1 << endl;
		}
		else if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "shininess" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(shininess);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shininess = " << shininess << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		TurbulenceTexture *tex = TurbulenceTexture::CreateTexture();
		tex->setScale( noiseScale );
		tex->setColor0( color0 );
		tex->setColor1( color1 );

		TexturedMaterial *newMat = TexturedMaterial::CreateMaterial();
		newMat->setAmbientCoeff( ambient );
		newMat->setPhongCoeff( specular );
		newMat->setShininess( shininess );
		newMat->setTexture( tex );
		*mat = newMat;
	}

	return m_ErrorFlag;
}

bool ConfigLoader::parseDturbulenceTexture( Tokenizer &tokenizer, Material **mat )
{
	m_ErrorFlag |= tokenizer.requireChar('{');

	vec3 ambient(0,0,0), specular(0,0,0);
	float shininess = 32.f;

	float noiseScale = 1.5f;
	float noiseDeriv = 2.f;
	vec3 color0(0.1, 0.1, 0.1), color1(0.0, 1.0, 0.0);

	while ( ! m_ErrorFlag && ! tokenizer.isEndOfFile() )
	{
		string token;
		//m_ErrorFlag |= tokenizer.readString(token);
		m_ErrorFlag |= tokenizer.readToken(token);

		if ( "noise_scale" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(noiseScale);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": noise_scale = " << noiseScale << endl;
		}
		else if ( "noise_derivative" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(noiseDeriv);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": noise_derivative = " << noiseDeriv << endl;
		}
		else if ( "color0" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color0);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color0 = " << color0 << endl;
		}
		else if ( "color1" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(color1);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": color1 = " << color1 << endl;
		}
		else if ( "ambient" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(ambient);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": ambient = " << ambient << endl;
		}
		else if ( "specular" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readVec3(specular);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": specular = " << specular << endl;
		}
		else if ( "shininess" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar('=');
			m_ErrorFlag |= tokenizer.readFloat(shininess);
			m_ErrorFlag |= tokenizer.requireChar(';');

			if (m_IsVerbose)
				cerr << __FUNCTION__ << ": shininess = " << shininess << endl;
		}
		else if ( "}" == token )
		{
			m_ErrorFlag |= tokenizer.requireChar(';');
			break;
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token 
				<< " at line " << tokenizer.getCurrentLineNumber() << endl;
			m_ErrorFlag = true;
			return true;
		}
	}

	if ( ! m_ErrorFlag)
	{
		DerivTurbulenceTexture *tex = DerivTurbulenceTexture::CreateTexture();
		tex->setScale( noiseScale );
		tex->setDeriv( noiseDeriv );
		tex->setColor0( color0 );
		tex->setColor1( color1 );

		TexturedMaterial *newMat = TexturedMaterial::CreateMaterial();
		newMat->setAmbientCoeff( ambient );
		newMat->setPhongCoeff( specular );
		newMat->setShininess( shininess );
		newMat->setTexture( tex );
		*mat = newMat;
	}

	return m_ErrorFlag;
}


