#pragma once

#include <vector>
#include <map>

#include "Tokenizer.h"

#include "GeometricObject.h"
#include "GeometryInstance.h"
#include "Material.h"
#include "LightSource.h"

#include "EnvironmentMap.h"

#include "MiscSettings.h"

class ConfigLoader
{
public:

	bool load(const char *filename, bool isVerbose = false);

private:
	static std::map<std::string, GeometricObject *> s_GeometryDictionary;
	static std::map<std::string, Material *> s_MaterialDictionary;

	std::vector<GeometricObject *> m_Objects;
	std::vector<LightSource *> m_LightSources;

	EnvironmentMap *m_pEnvironmentMap;

	bool m_ErrorFlag;
	bool m_IsVerbose;

	MiscSettings m_MiscSettings;

	static bool IsGeometryToken( const std::string &token );
	static bool IsTransformationToken( const std::string &token );
	static bool IsMaterialToken( const std::string &token );

	bool includeFile(const char *filename);

	bool parseGlobalConfig( Tokenizer &tokenizer );
	bool parsePixelFilter( Tokenizer &tokenizer );
	bool parseCamera( Tokenizer &tokenizer );
	bool parseScreen( Tokenizer &tokenizer );

	bool parsePointLight( Tokenizer &tokenizer );
	bool parseDirectionalLight( Tokenizer &tokenizer );
	bool parseEnvironmentLight( Tokenizer &tokenizer );

	bool parseGeometry( const std::string &token, Tokenizer &tokenizer, GeometricObject **obj );
	bool parseSphere( Tokenizer &tokenizer, GeometricObject **obj );
	bool parsePlane( Tokenizer &tokenizer, GeometricObject **obj );
	bool parseBox( Tokenizer &tokenizer, GeometricObject **obj );
	bool parseCylinder( Tokenizer &tokenizer, GeometricObject **obj );
	bool parseCone( Tokenizer &tokenizer, GeometricObject **obj );
	bool parseTriangle( Tokenizer &tokenizer, GeometricObject **obj );
	bool parseTriangleMesh( Tokenizer &tokenizer, GeometricObject **obj );
	bool parseMetaballSurface( Tokenizer &tokenizer, GeometricObject **obj );

	bool parseTransformation( const std::string &token, Tokenizer &tokenizer, GeometryInstance **obj );
	bool parseTransform( Tokenizer &tokenizer, GeometryInstance **obj );
	bool parseScale( Tokenizer &tokenizer, GeometryInstance **obj );
	bool parseRotate( Tokenizer &tokenizer, GeometryInstance **obj );
	bool parseTranslate( Tokenizer &tokenizer, GeometryInstance **obj);
	// WIgO[ṽT|[gǂ邩H 

	bool parseMaterial( const std::string &token, Tokenizer &tokenizer, Material **mat );
	bool parsePseudoNormalColor( Tokenizer &tokenizer, Material **mat );
	bool parseAmbient( Tokenizer &tokenizer, Material **mat );
	bool parseDiffuse( Tokenizer &tokenizer, Material **mat );
	bool parsePhong( Tokenizer &tokenizer, Material **mat );
	bool parseMirror( Tokenizer &tokenizer, Material **mat );
	bool parseRefract( Tokenizer &tokenizer, Material **mat );
	bool parseCheckerTexture( Tokenizer &tokenizer, Material **mat );
	bool parseSolidCheckerTexture( Tokenizer &tokenizer, Material **mat );
	bool parseNoiseTexture( Tokenizer &tokenizer, Material **mat );
	bool parseMarbleTexture( Tokenizer &tokenizer, Material **mat );
	bool parseWoodenTexture( Tokenizer &tokenizer, Material **mat );
	bool parseImageTexture( Tokenizer &tokenizer, Material **mat );
	bool parseTurbulenceTexture( Tokenizer &tokenizer, Material **mat );
	bool parseDturbulenceTexture( Tokenizer &tokenizer, Material **mat );
};





