#pragma once

#include "Texture.h"
#include "SolidNoise.h"

class DerivTurbulenceTexture : public Texture
{
private:
	DerivTurbulenceTexture()
		: scale(1.5f), octaves(5), deriv(2.f),
		color0(0.1f), color1(0.f, 1.f, 0.f)
	{
	}

public:
	static DerivTurbulenceTexture *CreateTexture()
	{
		DerivTurbulenceTexture *t = new DerivTurbulenceTexture();
		s_TextureCache.push_back( t );
		return t;
	}

	MyAlgebra::vec3 getColor(const MyAlgebra::vec3 &uvw) const
	{
		const MyAlgebra::Real t = (1.f + solid_noise.dturbulence(scale*uvw, octaves, deriv)) / 2.f;
		return t * color0 + (1.f-t) * color1;
	}

	float getScale() const { return scale; }
	void setScale(float s) { scale = s; }

	float getDeriv() const { return deriv; }
	void setDeriv(float d) { deriv = d; }

	void setColor0(const MyAlgebra::vec3 &c) { color0.set(c); }
	void setColor1(const MyAlgebra::vec3 &c) { color1.set(c); }
	void setColor0(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color0.set(r, g, b); }
	void setColor1(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color1.set(r, g, b); }

	MyAlgebra::vec3 getColor0() const { return color0; }
	MyAlgebra::vec3 getColor1() const { return color1; }

	//void clone(Texture **t) const { *t = new DerivTurbulenceTexture(*this); }

	Texture_Type getTextureType() const { return Dturbulence_Texture; }
	bool isSolidTexture() const { return true; }

private:
	float scale, deriv;
	int octaves;
	MyAlgebra::vec3 color0, color1;
	SolidNoise solid_noise;
};
