#include "EnvironmentLightGenerator.h"

#pragma comment( lib, "EnvmapSampler.lib" )

using namespace std;
using namespace MyAlgebra;

void EnvironmentLightGenerator::generate(int nSamples, const EnvironmentMap *envmap, vector<LightSource *> &lights)
{
	if ( ! envmap)
	{
		cerr << __FUNCTION__ << ": environment map is null" << endl;
		return;
	}

	const int imgHeight = 256;

	ImageRGBf longitudeLatitudeEnvmap;

	envmap->convertToLongitudeLatitudeFormat(imgHeight, longitudeLatitudeEnvmap);

	vector<vec3> directions, colors;

	EnvironmentMapSampler sampler;
	sampler.sample(nSamples, longitudeLatitudeEnvmap, directions, colors);

	for (int i=0; i<nSamples; i++)
	{
		//cerr << "sample[" << i << "] dir=" << directions[i] << ", color=" << colors[i] << endl;
 
		LightSource *l = new EnvironmentLight(-directions[i], colors[i]);
		lights.push_back( l );
	}
}

void EnvironmentLightGenerator::eraseOldSamples(vector<LightSource *> &lights)
{
	vector<LightSource *>::iterator itr = lights.begin();

	while (itr != lights.end())
	{
		if ((*itr)->getLightType() == LightSource::Environment_Light)
		{
			itr = lights.erase( itr );
		}
		else
		{
			++itr;
		}
	}
}

