#pragma once

#include "my_algebra.h"
#include <cstdlib>
#include <GL/glut.h>
#include "ImageRect.h"

class EnvironmentMap
{
public:
	EnvironmentMap()
		: m_pData(0), m_HdrGamma(0.5f)
	{
		m_SkyDomeTextures[0] = 0;
	}

	~EnvironmentMap()
	{
		if (m_pData) delete [] m_pData;
		if (m_SkyDomeTextures[0]) glDeleteTextures(6, m_SkyDomeTextures);
	}

	MyAlgebra::vec3 fetchColor(const MyAlgebra::vec3 &dir) const;

	bool load(const char *filename);

	void convertToLongitudeLatitudeFormat(int height, ImageRGBf &out) const;

	inline int getImageWidth() const { return m_ImageWidth; }
	inline int getImageHeight() const { return m_ImageHeight; }
	inline int getFaceWidth() const { return m_FaceWidth; }
	inline int getFaceHeight() const { return m_FaceHeight; }

	inline void setHDRGamma(float g) { m_HdrGamma = g; }
	inline float getHDRGamma() const { return m_HdrGamma; }

	bool isTextureReady() const { return m_SkyDomeTextures[0] > 0; }

	void drawSkyBox() const;

	void makeGLTextures();

private:
	int m_ImageWidth, m_ImageHeight;
	int m_FaceWidth, m_FaceHeight;
	MyAlgebra::vec3 *m_pData;	// rectangle, 6*face_width x face_height

	float m_HdrGamma;

	unsigned int m_SkyDomeTextures[6];

	bool loadFromCrossHDR(const char *filename);
	MyAlgebra::vec3 readTexture(const MyAlgebra::vec3 *addr, float u, float v) const;

	unsigned char *getPixelAddr(unsigned char *addr, int xi, int yi) const
	{
		return addr + 4 * (xi + m_ImageWidth * (m_ImageHeight - 1 - yi));
	}
};

