#pragma once

#include "Texture.h"
#include "SolidNoise.h"

class FlameTexture : public Texture
{
public:
	FlameTexture()
		: scale(5.f), octaves(2),
		color0(0, 1, 1), color1(1, 0.5, 0.5)
	{
		float stripes_per_unit = 0.15f;
		freq = (float)M_PI * stripes_per_unit;
	}

	MyAlgebra::vec3 getColor(const MyAlgebra::vec3 &uvw) const
	{
		const float temp = scale * 1.f / fabs( solid_noise.turbulence(freq*uvw, octaves) );
		const float t = 0.5f * sinf(temp) + 0.5f;

		return t * color0 + (1.f-t) * color1;
	}

	float getScale() const { return scale; }
	void setScale(float s) { scale = s; }
	void setStripesPerUnit(float p) { freq = (float)M_PI * p; }

	void setColor0(const MyAlgebra::vec3 &c) { color0.set(c); }
	void setColor1(const MyAlgebra::vec3 &c) { color1.set(c); }
	void setColor0(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color0.set(r, g, b); }
	void setColor1(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color1.set(r, g, b); }

	MyAlgebra::vec3 getColor0() const { return color0; }
	MyAlgebra::vec3 getColor1() const { return color1; }

	void clone(Texture **t) const { *t = new FlameTexture(*this); }

private:
	float freq, scale;
	int octaves;
	MyAlgebra::vec3 color0, color1;
	SolidNoise solid_noise;
};
