#pragma once

#include "my_algebra.h"
#include "Ray.h"
#include "Material.h"
#include "Texture.h"

#include <vector>

class GeometricObject
{
public:
	static void ClearGeometricObjectCache()
	{
		for (int i=0; i<(int)s_GeometricObjectCache.size(); i++)
			delete s_GeometricObjectCache[i];
		s_GeometricObjectCache.clear();
	}

	GeometricObject()
		: material(0)
	{
	}

	virtual ~GeometricObject()
	{
		// material deletion is managed in a static function of Material class
		//if (material) delete material;
	}

	virtual bool hit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const = 0;
	virtual bool shadowHit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const = 0;

	virtual void drawGL() const = 0;	// for preview using OpenGL

	Material *getMaterial() const { return material; }
	void setMaterial(Material *m) { material = m; }

	virtual MyAlgebra::vec3 getBoundingBoxMin() const { return MyAlgebra::vec3(0,0,0); }	// to be implemented
	virtual MyAlgebra::vec3 getBoundingBoxMax() const { return MyAlgebra::vec3(0,0,0); }	// to be implemented

protected:
	static std::vector<GeometricObject *> s_GeometricObjectCache;

	Material *material;

};

