#pragma once

#include "GeometricObject.h"
#include "my_algebra.h"

class GeometryInstance : public GeometricObject
{
private:
	GeometryInstance()
		: m_pObject(0), m_InverseTransformMatrix(MyAlgebra::mat4_id)
	{
	}

	GeometryInstance(GeometricObject *o)
		: m_pObject(o), m_InverseTransformMatrix(MyAlgebra::mat4_id)
	{
		material = o->getMaterial();
	}

public:
	static GeometryInstance *CreateGeometricObject()
	{
		GeometryInstance *obj = new GeometryInstance();
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static GeometryInstance *CreateGeometricObject(GeometricObject *o)
	{
		GeometryInstance *obj = new GeometryInstance(o);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	bool hit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const;
	bool shadowHit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const;

	void drawGL() const;	// for preview using OpenGL

	GeometricObject *getObject() { return m_pObject; }
	void setObject(GeometricObject *o)
	{
		m_pObject = o;
		if ( ! material)
			material = o->getMaterial();
	}

	GeometryInstance &transform(const MyAlgebra::vec4 &col0, const MyAlgebra::vec4 &col1, const MyAlgebra::vec4 &col2, const MyAlgebra::vec4 &col3);
	GeometryInstance &transform(const MyAlgebra::mat4 &M);

	GeometryInstance &translate(float x, float y, float z);
	GeometryInstance &translate(const MyAlgebra::vec3 &t);

	GeometryInstance &scale(float sx, float sy, float sz);

	GeometryInstance &rotate(float angle, float x, float y, float z);

private:
	GeometricObject *m_pObject;

	MyAlgebra::mat4 m_InverseTransformMatrix;

};
