#pragma once

#include "Texture.h"
#include "ImageRect.h"
#include "ImageIO.h"
#include <string>

class ImageTexture : public Texture
{
private:
	ImageTexture()
	{
	}

	ImageTexture(const ImageTexture &tex)
		: image(tex.image), Texture(tex)
	{
	}

public:
	static ImageTexture *CreateTexture()
	{
		ImageTexture *t = new ImageTexture();
		s_TextureCache.push_back( t );
		return t;
	}

	static ImageTexture *CreateTexture(const ImageTexture *_t)
	{
		ImageTexture *t = new ImageTexture(*_t);
		s_TextureCache.push_back( t );
		return t;
	}

	~ImageTexture()
	{
	}

	MyAlgebra::vec3 getColor(const MyAlgebra::vec3 &uvw) const
	{
		if ( ! image.getData()) return MyAlgebra::vec3(0.f);

		const float x = uvw.x * image.getWidth();
		const float y = uvw.y * image.getHeight();
		return image.bilinearInterp(x, y);
	}

	bool loadImage(const std::string &filename)
	{
		ImageIO io;
		return io.load(filename.c_str(), image);
	}

	Texture_Type getTextureType() const { return Image_Texture; }
	bool isSolidTexture() const { return false; }

private:
	ImageRGBf image;
};