#pragma once

#include "Texture.h"
#include "SolidNoise.h"

class MarbleTexture : public Texture
{
private:
	MarbleTexture()
		: scale(5.f), octaves(8),
		color0(0.8f), color1(0.4f, 0.2f, 0.1f), color2(0.06f,0.04f,0.02f)
	{
		float stripes_per_unit = 0.15f;
		freq = (float)M_PI * stripes_per_unit;
	}

public:
	static MarbleTexture *CreateTexture()
	{
		MarbleTexture *t = new MarbleTexture();
		s_TextureCache.push_back( t );
		return t;
	}

	MyAlgebra::vec3 getColor(const MyAlgebra::vec3 &uvw) const
	{
		const float temp = scale * solid_noise.turbulence(freq*uvw, octaves);
		float t = 2.f * fabs(sinf(freq*uvw.x + temp));

		if (t < 1.f)
		{
			return t * color1 + (1.f-t) * color2;
		}
		else
		{
			t -= 1.f;
			return t * color0 + (1.f-t) * color1;
		}
	}

	float getScale() const { return scale; }
	void setScale(float s) { scale = s; }
	void setStripesPerUnit(float p) { freq = (float)M_PI * p; }

	void setColor0(const MyAlgebra::vec3 &c) { color0.set(c); }
	void setColor1(const MyAlgebra::vec3 &c) { color1.set(c); }
	void setColor2(const MyAlgebra::vec3 &c) { color2.set(c); }
	void setColor0(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color0.set(r, g, b); }
	void setColor1(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color1.set(r, g, b); }
	void setColor2(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color2.set(r, g, b); }

	MyAlgebra::vec3 getColor0() const { return color0; }
	MyAlgebra::vec3 getColor1() const { return color1; }
	MyAlgebra::vec3 getColor2() const { return color2; }

	Texture_Type getTextureType() const { return Marble_Texture; }
	bool isSolidTexture() const { return true; }

private:
	float freq, scale;
	int octaves;
	MyAlgebra::vec3 color0, color1, color2;
	SolidNoise solid_noise;
};
