#pragma once

#include "../my_algebra.h"
#include <vector>

struct ObjMaterial
{
	std::string mat_name;
	MyAlgebra::vec3 ambient;
	MyAlgebra::vec3 diffuse;
	MyAlgebra::vec3 specular;
	float alpha;
	float shininess;
	std::string tex_name;

	ObjMaterial()
		: ambient(0.2f,0.2f,0.2f),
		diffuse(0.8f,0.8f,0.8f),
		specular(1.f,1.f,1.f),
		alpha(1.f),
		shininess(0.f)
	{
	}
};

class ObjMesh
{
public:
	typedef MyAlgebra::vec3 vec3;
	typedef MyAlgebra::vec2 vec2;

	ObjMesh()
	{
	}

	~ObjMesh()
	{
	}

	inline void setGroupName( const std::string &name )
	{
		group_name = name;
	}

	inline std::string getGroupName() const { return group_name; }

	inline void setUseMaterialName( const std::string &name )
	{
		use_material_name = name;
	}

	inline std::string getUseMaterialName() const { return use_material_name; }

	inline int getNumVertices() const { return (int)vertices.size(); }

	inline void addVertex( const vec3 &p )
	{
		vertices.push_back( p );
	}

	inline vec3 &getVertex(const int i) { return vertices[i]; }
	inline std::vector<vec3> &getVertices() { return vertices; }

	inline void addNormal( const MyAlgebra::vec3 &n )
	{
		normals.push_back( n );
	}

	inline int getNumNormals() const { return (int)normals.size(); }

	inline vec3 &getNormal(const int i) { return normals[i]; }
	inline std::vector<vec3> &getNormals() { return normals; }

	inline void addTexcoord( const MyAlgebra::vec2 &t )
	{
		texcoords.push_back( t );
	}

	inline int getNumTexCoords() const { return (int)texcoords.size(); }

	inline vec2 &getTexcoord(const int i) { return texcoords[i]; }
	inline std::vector<vec2> &getTexcoords() { return texcoords; }

	inline int getNumFaces() const { return (int)face_vertices.size(); }

	inline void addFaceVertex( const std::vector<int> &v )
	{
		face_vertices.push_back( v );
	}

	inline std::vector<int> &getFaceVertices(const int i) { return face_vertices[i]; }

	inline void addFaceNormal( const std::vector<int> &n )
	{
		face_normals.push_back( n );
	}

	inline std::vector<int> &getFaceNormals(const int i) { return face_normals[i]; }

	inline void addFaceTexcoord( const std::vector<int> &t )
	{
		face_texcoords.push_back( t );
	}

	inline std::vector<int> &getFaceTexCoord(const int i) { return face_texcoords[i]; }

	// For material

	inline void setMaterialName(const std::string& name) { mat.mat_name = name; }
	inline const std::string& getMaterialName() const { return mat.mat_name; }

	inline void setAmbient(const vec3& color) { mat.ambient = color; }
	inline const vec3& getAmbient() const { return mat.ambient; }

	inline void setDiffuse(const vec3& color) { mat.diffuse = color; }
	inline const vec3& getDiffuse() const { return mat.diffuse; }

	inline void setSpecular(const vec3& color) { mat.specular = color; }
	inline const vec3& getSpecular() const { return mat.specular; }

	inline void setAlpha(float alpha) { mat.alpha = alpha; }
	inline float getAlpha() const { return mat.alpha; }

	inline void setShininess(float shininess) { mat.shininess = shininess; }
	inline float getShininess() const { return mat.shininess; }

	inline void setTexName(const std::string& name) { mat.tex_name = name; }
	inline const std::string& getTexName() const { return mat.tex_name; }

	// For debugging

	void dump()
	{
		std::cout << "vertices:" << std::endl;

		for (int i=0; i<(int)vertices.size(); i++)
		{
			std::cout << vertices[i] << std::endl;
		}

		std::cout << "face vertices:" << std::endl;

		for (int i=0; i<(int)face_vertices.size(); i++)
		{
			for (int j=0; j<(int)face_vertices[i].size(); j++)
			{
				std::cout << face_vertices[i][j] << ", ";
			}
			std::cout << std::endl;
		}

		std::cout << "face texcoords:" << std::endl;

		for (int i=0; i<(int)face_texcoords.size(); i++)
		{
			for (int j=0; j<(int)face_texcoords[i].size(); j++)
			{
				std::cout << face_texcoords[i][j] << ", ";
			}
			std::cout << std::endl;
		}

		std::cout << "face normals:" << std::endl;

		for (int i=0; i<(int)face_normals.size(); i++)
		{
			for (int j=0; j<(int)face_normals[i].size(); j++)
			{
				std::cout << face_normals[i][j] << ", ";
			}
			std::cout << std::endl;
		}
	}


private:
	std::string group_name;

	std::string use_material_name;

	ObjMaterial mat;

	std::vector< vec3 > vertices;
	std::vector< vec3 > normals;
	std::vector< vec2 > texcoords;

	std::vector< std::vector<int> > face_vertices;
	std::vector< std::vector<int> > face_normals;
	std::vector< std::vector<int> > face_texcoords;
};
