#include "PlyReader.h"
#include <iostream>
#include <list>

using namespace std;
using namespace MyAlgebra;

namespace PlyReader
{
	// Variables

	vec3 *positions = 0;
	vec3 *normals = 0;
	unsigned char *colors = 0;
	unsigned int *indices = 0;

	int n_vertices = 0;
	int n_triangles = 0;

	bool normal_available = false;
	bool color_available = false;

	int vertex_index = 0;
	int vertex_property_index = 0;

	int normal_index = 0;
	int normal_property_index = 0;

	int color_index = 0;
	int color_property_index = 0;

	int triangle_index = 0;

	// Functions

	int getNumVertices() { return n_vertices; }
	int getNumTriangles() { return n_triangles; }

	bool isColorAvailable() { return color_available; }

	MyAlgebra::vec3 *getVertices() { return positions; }
	MyAlgebra::vec3 *getNormals() { return normals; }
	unsigned char *getColors() { return colors; }
	unsigned int *getIndices() { return indices; }

	void clean_up()
	{
		n_vertices = n_triangles = 0;

		if (positions) {
			delete [] positions; positions = 0;
		}

		if (normals) {
			delete [] normals; normals = 0;
		}

		if (colors) {
			delete [] colors; colors = 0;
		}

		if (indices) {
			delete [] indices; indices = 0;
		}

		normal_available = color_available = false;

		vertex_index = vertex_property_index = 0;
		normal_index = normal_property_index = 0;
		color_index = color_property_index = 0;
		triangle_index = 0;
	}

	int vertex_cb(p_ply_argument argument)
	{
		long eol;
		ply_get_argument_user_data(argument, 0, &eol);

		if (vertex_property_index < 3)
			positions[vertex_index][vertex_property_index] = (float)ply_get_argument_value(argument);
	
		//cerr << "vertex_cb[" << element_index << "]: " << property_index << endl;

		if (eol) {
			vertex_index++;
			vertex_property_index = 0;
		} else {
			vertex_property_index++;
		}

		return 1;
	}

	int normal_cb(p_ply_argument argument)
	{
		long eol;
		ply_get_argument_user_data(argument, 0, &eol);

		if (normal_property_index < 3)
			normals[normal_index][normal_property_index] = (float)ply_get_argument_value(argument);

		//cerr << "normal_cb[" << element_index << "]: " << property_index << endl;

		if (eol) {
			normal_index++;
			normal_property_index = 0;
		} else {
			normal_property_index++;
		}

		return 1;
	}

	int color_cb(p_ply_argument argument)
	{
		long eol;
		ply_get_argument_user_data(argument, 0, &eol);

		if (color_property_index < 3)
			colors[3 * color_index + color_property_index] = (unsigned char)ply_get_argument_value(argument);

		//cerr << "color_cb[" << element_index << "]: " << property_index << endl;

		if (eol) {
			color_index++;
			color_property_index = 0;
		} else {
			color_property_index++;
		}

		return 1;
	}

	int triangle_cb(p_ply_argument argument)
	{
		long length, value_index;
		ply_get_argument_property(argument, 0, &length, &value_index);

		if (value_index < 0) return 1;

		if (value_index > 2)
		{
			cerr << __FUNCTION__ << ": not triangle" << endl;
			clean_up();
			cerr << "Push <ENTER> to exit"; getchar(); exit(-1); 
		}

		indices[triangle_index++] = (int)ply_get_argument_value(argument);
		
		return 1;
	}

#if 0
	void computeNormals()
	{
		vec3 *face_normals = new vec3[ n_triangles ];
		std::list<int> *vertex_face = new std::list<int>[ n_triangles ];
		
		// computing face normals and areas
		for (int tri=0; tri<n_triangles; tri++)
		{
			const int i1 = indices[3*tri  ];
			const int i2 = indices[3*tri+1];
			const int i3 = indices[3*tri+2];

			const vec3& v1 = positions[ i1 ];
			const vec3& v2 = positions[ i2 ];
			const vec3& v3 = positions[ i3 ];
			
			vertex_face[ i1 ].push_back( tri );
			vertex_face[ i2 ].push_back( tri );
			vertex_face[ i3 ].push_back( tri );

			// ʐςŏdݕtꂽPʖ@xNg
			face_normals[ tri ] = (v2 - v1) ^ (v3 - v2);
		}
		
		// computing vertex normals by averaging face normals
		for (int v=0; v<n_vertices; v++)
		{
			vec3 vertex_normal(0,0,0);

			std::list<int>::iterator itr = vertex_face[ v ].begin();
			while (itr != vertex_face[ v ].end())
			{
				const int tri = (int)*itr;
				vertex_normal += face_normals[ tri ];
				++itr;
			}

			vertex_normal.normalize();
			normals[ v ] = vertex_normal;
		}

		delete [] face_normals;
		delete [] vertex_face;
	}
#endif

	bool read(const char *filename)
	{
		clean_up();

		p_ply iply = ply_open(filename, NULL);
		if (! iply)
		{
			cerr << __FUNCTION__ << ": cannot open file: " << filename << endl;
			return false;
		}

		if (! ply_read_header(iply))
		{
			cerr << __FUNCTION__ << ": cannot read header: " << filename << endl;
			return false;
		}

		cerr << __FUNCTION__ << ": " << filename << endl;

		p_ply_element element = 0;
		while ((element = ply_get_next_element(iply, element)))
		{
			long n_instances = 0;
			const char *element_name;
			
			ply_get_element_info(element, &element_name, &n_instances);

			if (! strcmp("vertex", element_name) )
			{
				n_vertices = n_instances;
			}
			else if (! strcmp("face", element_name) )
			{
				n_triangles = ply_set_read_cb(iply, "face", "vertex_indices", triangle_cb, 0, 0);
				continue;
			}
			else
			{
				cerr << __FUNCTION__ << ": ignoring element: " << element_name << endl;
				continue;
			}

			p_ply_property property = 0;
			while ((property = ply_get_next_property(element, property)))
			{
				const char *property_name;
				e_ply_type type, length_type, value_type;

				ply_get_property_info(property, &property_name, &type, &length_type, &value_type);

				if ( 'x' == property_name[0])
				{
					ply_set_read_cb(iply, "vertex", "x", vertex_cb, 0, 0);
					ply_set_read_cb(iply, "vertex", "y", vertex_cb, 0, 0);
					ply_set_read_cb(iply, "vertex", "z", vertex_cb, 0, 1);
				}
				else if ( 'y' == property_name[0] || 'z' == property_name[0])
				{
					continue;
				}
				else if (! strcmp("nx", property_name) )
				{
					normal_available = true;

					ply_set_read_cb(iply, "vertex", "nx", normal_cb, 0, 0);
					ply_set_read_cb(iply, "vertex", "ny", normal_cb, 0, 0);
					ply_set_read_cb(iply, "vertex", "nz", normal_cb, 0, 1);
				}
				else if (! strcmp("ny", property_name) || ! strcmp("nz", property_name))
				{
					continue;
				}
				else if (! strcmp("red", property_name) )
				{
					if (type != PLY_UCHAR)
					{
						cerr << __FUNCTION__ << ": color type is not unsigned char" << endl;
						continue;
					}

					color_available = true;

					ply_set_read_cb(iply, "vertex", "red", color_cb, 0, 0);
					ply_set_read_cb(iply, "vertex", "green", color_cb, 0, 0);
					ply_set_read_cb(iply, "vertex", "blue", color_cb, 0, 1);
				}
				else if (! strcmp("green", property_name) || ! strcmp("blue", property_name) )
				{
					continue;
				}
				else
				{
					cerr << __FUNCTION__ << ": ignoring property: " << element_name << "::" << property_name << endl;
					continue;
				}
			}
		}
		
		if (! n_vertices)
		{
			cerr << __FUNCTION__ << ": no vertices" << endl;
			ply_close( iply );
			return false;
		}
		else if (! n_triangles)
		{
			cerr << __FUNCTION__ << ": no triangles" << endl;
			ply_close( iply );
			return false;
		}
		
		cerr << "  # of vertices: " << n_vertices << endl
			 << "  # of triangles: " << n_triangles << endl
			 << "  color: " << (PlyReader::isColorAvailable() ? "available" : "unavailable") << endl;

		positions = new vec3[ n_vertices ];
		normals = new vec3[ n_vertices ];
		if (color_available) colors = new unsigned char[ 3 * n_vertices ];
		indices = new unsigned int[ 3 * n_triangles ];
		
		if (! ply_read(iply))
		{
			cerr << __FUNCTION__ << ": cannot read data" << endl;
			ply_close( iply );
			clean_up();
			return false;
		}

		ply_close( iply );

		if (! normal_available)
		{
			cerr << __FUNCTION__ << ": computing normals" << endl;
			//computeNormals();
		}

		return true;
	}
}
