#pragma once

#include "GeometricObject.h"
#include <vector>
#include "my_algebra.h"

class Metaball
{
public:
	Metaball()
	{
	}

	Metaball(const MyAlgebra::vec3 &c, float r)
		: m_Center(c), m_Radius(r)
	{
	}

	Metaball(const Metaball &m)
		: m_Center(m.m_Center), m_Radius(m.m_Radius)
	{
	}

	const MyAlgebra::vec3 &getCenter() const { return m_Center; }
	MyAlgebra::vec3 &getCenter() { return m_Center; }
	void setCenter(const MyAlgebra::vec3 &c) { m_Center = c; }

	float getRadius() const { return m_Radius; }
	void setRadius(float r) { m_Radius = r; }

private:
	MyAlgebra::vec3 m_Center;
	float m_Radius;

};

class MetaballSurface : public GeometricObject
{
private:
	MetaballSurface()
	{
	}

	MetaballSurface(const MetaballSurface &s)
		: m_Metaballs(s.m_Metaballs)
	{
		material = s.material;
	}

public:
	static MetaballSurface *CreateGeometricObject()
	{
		MetaballSurface *obj = new MetaballSurface();
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static MetaballSurface *CreateGeometricObject(const MetaballSurface &s)
	{
		MetaballSurface *obj = new MetaballSurface(s);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	bool hit(const Ray &ray, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const;
	bool shadowHit(const Ray &ray, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const;

	void drawGL() const;	// for preview using OpenGL

	void clearMetaballs() { m_Metaballs.clear(); }

	int getNumMetaballs() const { return (int)m_Metaballs.size(); }

	const std::vector<Metaball> &getMetaballs() const { return m_Metaballs; }
	std::vector<Metaball> &getMetaballs() { return m_Metaballs; }
	const Metaball &getMetaball(int i) const { return m_Metaballs[i]; }
	Metaball &getMetaball(int i) { return m_Metaballs[i]; }

	void addMetaball(const Metaball &m) { m_Metaballs.push_back( m ); }

private:
	std::vector<Metaball> m_Metaballs;

	void calculateMetaballBezierCoeffs(float intensity, float a, float t0, float t1, float *coeffs) const;
};



