#pragma once

#include "Texture.h"
#include "SolidNoise.h"

// from "Realistic Ray Tracing, second ed.", p. 85

class NoiseTexture : public Texture
{
private:
	NoiseTexture()
		: color0(0.8f,0.f,0.f), color1(0.f,0.f,0.8f), scale(1.f)
	{
	}

	NoiseTexture(const MyAlgebra::vec3 &c1, const MyAlgebra::vec3 &c2, float _scale=1.f)
		: color0(c1), color1(c2), scale(_scale)
	{
	}

public:
	static NoiseTexture *CreateTexture()
	{
		NoiseTexture *t = new NoiseTexture();
		s_TextureCache.push_back( t );
		return t;
	}

	static NoiseTexture *CreateTexture(const MyAlgebra::vec3 &c1, const MyAlgebra::vec3 &c2, float _scale=1.f)
	{
		NoiseTexture *t = new NoiseTexture(c1, c2, _scale);
		s_TextureCache.push_back( t );
		return t;
	}

	MyAlgebra::vec3 getColor(const MyAlgebra::vec3 &uvw) const
	{
		const MyAlgebra::Real t = (1.f + solid_noise.noise(uvw*scale)) / 2.f;
		return t * color0 + (1.f-t) * color1;
	}

	float getScale() const { return scale; }
	void setScale(float s) { scale = s; }

	void setColor0(const MyAlgebra::vec3 &c) { color0.set(c); }
	void setColor1(const MyAlgebra::vec3 &c) { color1.set(c); }
	void setColor0(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color0.set(r, g, b); }
	void setColor1(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color1.set(r, g, b); }

	MyAlgebra::vec3 getColor0() const { return color0; }
	MyAlgebra::vec3 getColor1() const { return color1; }

	Texture_Type getTextureType() const { return Noise_Texture; }
	bool isSolidTexture() const { return true; }

private:
	float scale;
	MyAlgebra::vec3 color0, color1;
	SolidNoise solid_noise;
};
