#pragma once

#include <string>
#include <vector>
#include <cstdlib>
#include <GL/glut.h>
#include <iostream>

class OGLBaseWindow
{
public:
	static void Display();
	static void Reshape(int w, int h);
	static void Keyboard(unsigned char key, int x, int y);
	static void Special(int key, int x, int _y);
	static void Mouse(int button, int state, int x, int _y);
	static void Motion(int x, int _y);
	static void Passive(int x, int _y);
	static void Menu(int item);
	static void Idle();

	OGLBaseWindow()
		: m_WindowID(0), m_WindowWidth(256), m_WindowHeight(256)
	{
	}

	virtual ~OGLBaseWindow()
	{
		for (unsigned int i=0; i<s_WindowsArray.size(); i++)
		{
			if ( s_WindowsArray[i] == this )
			{
				s_WindowsArray.erase( s_WindowsArray.begin() + i );
				break;
			}
		}

		glutDestroyWindow(m_WindowID);
	}

	inline int getWindowID() const { return m_WindowID; }
	const std::string &getWindowName() const { return m_WindowName; }

	inline int getWidth() const { return m_WindowWidth; }
	inline int getHeight() const { return m_WindowHeight; }

	inline void setInitialWindowSize(int w, int h)
	{
		m_WindowWidth = w;
		m_WindowHeight = h;
	}

	inline void setInitialWindowPos(int x, int y)
	{
		glutInitWindowPosition(x, y);
	}

	inline void hideWindow() const
	{
		glutSetWindow(m_WindowID);
		glutHideWindow();
	}

	inline void showWindow() const
	{
		glutSetWindow(m_WindowID);
		glutShowWindow();
	}

	inline void setWindowTitle(const char *title)
	{
		m_WindowName = title;
		glutSetWindow(m_WindowID);
		glutSetWindowTitle(title);
	}

	inline void create(const char *windowTitle)
	{
		glutInitWindowSize(m_WindowWidth, m_WindowHeight);

		m_WindowName = windowTitle;
		m_WindowID = glutCreateWindow(windowTitle);

		glutSetWindow(m_WindowID);
		glutDisplayFunc(Display);
		glutReshapeFunc(Reshape);
		glutKeyboardFunc(Keyboard);
		glutSpecialFunc(Special);
		glutMouseFunc(Mouse);
		glutMotionFunc(Motion);
		glutPassiveMotionFunc(Passive);

		s_WindowsArray.push_back( this );

		init();
	}

	// in subclasses, re-define the following functions if needed

	virtual void display() {}

	virtual void reshape(int w, int h)
	{
		if (h <= 1) return;
		m_WindowWidth = w;
		m_WindowHeight = h;
		glViewport(0, 0, w, h);
		gluOrtho2D(0, w, 0, h);
	}

	virtual void keyboard(unsigned char key, int x, int y) {}
	virtual void special(int key, int x, int y) {}
	virtual void mouse(int button, int state, int x, int _y) {}
	virtual void motion(int x, int _y) {}
	virtual void passive(int x, int _y) {}
	virtual void menu(int item) {}
	virtual void init() {}

protected:
	static std::vector<OGLBaseWindow *> s_WindowsArray;

	std::string m_WindowName;
	int m_WindowID;
	int m_WindowWidth, m_WindowHeight;
};
