#include "Plane.h"
#include <cstdlib>
#include <GL/glut.h>

using namespace std;
using namespace MyAlgebra;

// plane is defined such that <x, n> = d, where x is a 3D point, n is the normal
bool Plane::hit(const Ray &r, Real tmin, Real tmax, HitRecord &record) const
{
	const vec3 dir = r.getUnitDir();

	const Real a = dot(dir, normal);

	if (fabs(a) < 1.0e-10) return false;

	const Real b = dot(r.getOrigin(), normal);

	const Real t = -(b - d) / a;

	if (t < tmin || t > tmax) return false;

	record.t = t;
	record.normal = normal;
	record.hit_pos = r.getPosition(t);
	record.tex_coords.set(0.f);
	record.material = material;

	return true;
}

bool Plane::shadowHit(const Ray &r, Real tmin, Real tmax) const
{
	const vec3 dir = r.getUnitDir();

	const Real a = dot(dir, normal);

	if (fabs(a) < 1.0e-10) return false;

	const Real b = dot(r.getOrigin(), normal);

	const Real t = -(b - d) / a;

	if (t < tmin || t > tmax) return false;

	return true;
}

// for preview using OpenGL
void Plane::drawGL() const
{
	vec3 u,v;
	normal.getTwoOrthogonals(u, v);

	const Real s = 1.0e+5;
	u.scale( s );
	v.scale( s );

	const vec3 p = d * normal;

	glBegin(GL_QUADS);
	glVertex3fv( p - u - v );
	glVertex3fv( p + u - v );
	glVertex3fv( p + u + v );
	glVertex3fv( p - u + v );
	glEnd();

}
