#pragma once

#include "GeometricObject.h"

class Plane : public GeometricObject
{
private:
	Plane()
	{
	}

	Plane(const Plane &p)
		: normal(p.normal), d(p.d)
	{
		//p.material->clone(&material);
		material = p.material;
	}

	Plane(const MyAlgebra::vec3 &n, MyAlgebra::Real _d, Material *m)
		: normal(n), d(_d)
	{
		material = m;
	}

public:
	static Plane *CreateGeometricObject()
	{
		Plane *obj = new Plane();
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static Plane *CreateGeometricObject(const Plane &p)
	{
		Plane *obj = new Plane(p);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static Plane *CreateGeometricObject(const MyAlgebra::vec3 &n, MyAlgebra::Real _d, Material *m)
	{
		Plane *obj = new Plane(n, _d, m);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	bool hit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const;
	bool shadowHit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const;

	void drawGL() const;	// for preview using OpenGL

	inline MyAlgebra::vec3 getNormal() const { return normal; }
	inline MyAlgebra::Real getD() const { return d; }

	inline void setNormal(const MyAlgebra::vec3 &n) { normal = n; }
	inline void setD(MyAlgebra::Real _d) { d = _d; }

private:
	MyAlgebra::vec3 normal;
	MyAlgebra::Real d;

};

