#pragma once

#include "my_algebra.h"

class Ray
{
public:
	Ray()
	{
	}

	Ray(const Ray& r)
		: m_Origin(r.m_Origin), m_UnitDir(r.m_UnitDir)
	{
	}

	Ray(const MyAlgebra::vec3 &o, const MyAlgebra::vec3 &d)
		: m_Origin(o), m_UnitDir(d)
	{
	}

	~Ray()
	{
	}

	inline MyAlgebra::vec3 getOrigin() const { return m_Origin; }
	inline MyAlgebra::vec3 getUnitDir() const { return m_UnitDir; }

	inline MyAlgebra::vec3 getPosition(MyAlgebra::Real t) const { return m_Origin + t*m_UnitDir; }

	inline void setOrigin(const MyAlgebra::vec3 &o) { m_Origin = o; }
	inline void setUnitDir(const MyAlgebra::vec3 &d) { m_UnitDir = d; }
	inline void setDirWithNormalization(const MyAlgebra::vec3 &d)
	{
		m_UnitDir = d;
		m_UnitDir.normalize();
	}

	friend std::ostream &operator<<(std::ostream &os, const Ray &r)
	{
		os << r.getOrigin() << " + t" << r.getUnitDir();
		return os;
	}


private:
	MyAlgebra::vec3 m_Origin;
	MyAlgebra::vec3 m_UnitDir;	// should be a unit vector
};

