#include "RayTracer.h"
#include "Scene.h"
#include "Camera.h"
#include "Sampler.h"
#include "Filter.h"

#include "stopwatch.h"

using namespace std;
using namespace MyAlgebra;

extern int num_samples_per_pixel;

extern int dump_computational_time;
extern int current_xi;
extern int current_yi;

void RayTracer::renderScene(ImageRGBf &imageBuffer)
{
	Stopwatch watch;

	if (dump_computational_time)
		watch.Start();

	// get axes of camera coordinate system
	vec3 u, v, w;
	camera->getUVW(u, v, w);

	const vec3 eye = camera->getEyePosition();

	const int width = imageBuffer.getWidth();
	const int height = imageBuffer.getHeight();

	const Real screenDist = camera->calcScreenDistance((Real)height);

	const Real halfWidth = 0.5f * width;
	const Real halfHeight = 0.5f * height;

	vec2 *pixelOffsets = new vec2[num_samples_per_pixel];
	vec3 *samples = new vec3[num_samples_per_pixel];

	for (int yi=0; yi<height; yi++)
	{
		current_yi = yi;	// for debugging

		for (int xi=0; xi<width; xi++)
		{
			current_xi = xi;	// for debugging

			// assume that the provided samples are within [0, 1]^2
			image_sampler->sample(pixelOffsets, num_samples_per_pixel);

			for (int si=0; si<num_samples_per_pixel; si++)
			{
				// setup viewing ray
				vec3 dir = (xi+pixelOffsets[si].x-halfWidth)*u
						 + (yi+pixelOffsets[si].y-halfHeight)*v
						 - screenDist*w;
				dir.normalize();

				samples[si] = trace( Ray(eye, dir) );
			}

			imageBuffer(xi, yi) = image_filter->filter(samples, pixelOffsets, num_samples_per_pixel);
		}
	}

	delete [] pixelOffsets;
	delete [] samples;

	if (dump_computational_time)
		cout << __FUNCTION__ << ": time: " << watch.GetTime() << " sec" << endl;
}
