#pragma once

#ifdef USE_COUNTER
#include <vector>
extern std::vector<int> iterations_for_not_found;
extern std::vector<int> iterations_for_found;
extern std::vector<int> max_stack_counts;
#endif


//const float clipping_epsilon = 1.0e-2f;
//const float clipping_epsilon = 1.0e-3f;
const float clipping_epsilon = 1.0e-4f;
//const float clipping_epsilon = 1.0e-10f;
//const float clipping_epsilon = 0.015f;
//const float clipping_epsilon = 0.03f;
//const int num_loop_limit = 50;
//const int num_loop_limit = 30;
const int num_loop_limit = 20;
//const int num_newton_iterations = 3;
//const int num_newton_iterations = 2;
//const int num_newton_iterations = 1;
const int num_newton_iterations = 0;

const int polygon_stack_size = 10;
//const int polygon_stack_size = 6;
//const int polygon_stack_size = 3;

// Based on original, recursive Bezier Clipping

template <int degree>
class BezierClipping
{
public:
	BezierClipping()
	{
	}

	~BezierClipping()
	{
	}

	inline void initWithBezierCoeffs(const float *bezierCoeffs)
	{
		memcpy(initial_coeffs, bezierCoeffs, sizeof(float)*(degree+1));
	}

	inline void initWithPolyCoeffs(const float *polyCoeffs)
	{
		polynomial2BezierForm(polyCoeffs, initial_coeffs)
	}

	float findSmallestRoot() const
	{
		const int nInitialIntersections = countPolygonIntersections( initial_coeffs );
		if ( nInitialIntersections == 0) return -1.f;

		CtrlPolygon polygon_stack[polygon_stack_size];

		memcpy(polygon_stack[0].yCoords, initial_coeffs, sizeof(float)*(degree+1));
		polygon_stack[0].xMin = 0.f;
		polygon_stack[0].xMax = 1.f;
		polygon_stack[0].nIntersections = nInitialIntersections;

		int stack_index = 0;

	#ifdef USE_COUNTER
		int max_stack_count = 0;
	#endif

		for (int counter=0; counter<num_loop_limit; counter++)
		{
			CtrlPolygon& poly = polygon_stack[stack_index];

			const int nIntersections = poly.nIntersections;

			if ( nIntersections == 1 )
			{
				const float xMin = findXMin( poly );
				const float xMax = findXMax( poly );

				const float oldInterval = poly.xMax - poly.xMin;
				const float newInterval = xMax - xMin;

				if (newInterval <= clipping_epsilon)
				{
					const float root = 0.5f * (xMax + xMin);

					//for (int i=0; i<num_newton_iterations; i++)
					//	root = Newton(root);

				#ifdef USE_COUNTER
					iterations_for_found.push_back( counter );
					max_stack_counts.push_back( max_stack_count+1 );
				#endif

					return root;
				}
				else if (oldInterval - newInterval < 0.2f * oldInterval)
				{
					CtrlPolygon newLeft;
					memcpy(newLeft.yCoords, poly.yCoords, sizeof(float)*(degree+1));
					
					for (int i=1; i<=degree; i++)
						for (int j=degree; j>=i; j--)
							newLeft.yCoords[j] += (newLeft.yCoords[j-1] - newLeft.yCoords[j]) * 0.5f;

					for (int i=1; i<=degree; i++)
						for (int j=0; j<=degree-i; j++)
							poly.yCoords[j] += (poly.yCoords[j+1] - poly.yCoords[j]) * 0.5f;

					// poly  newLeft  {xMin, xMax} ̍XV
					const float divX = poly.xMin + oldInterval * 0.5f;
					newLeft.xMin = poly.xMin;
					newLeft.xMax = divX;
					poly.xMin = divX;

					// check # of intersections ... <1, 0> or <0, 1> or <0, 0>

					newLeft.nIntersections = countPolygonIntersections(newLeft.yCoords);
					if ( newLeft.nIntersections ) // _͌㔼ɂcO͕sv
					{
						polygon_stack[stack_index] = newLeft;
					}
					else if ( ! countPolygonIntersections(poly.yCoords) ) // ǂɂ_Ȃ
					{
						stack_index--;
					}

					if (stack_index < 0)
					{
					#ifdef USE_COUNTER
						iterations_for_not_found.push_back( counter );
						max_stack_counts.push_back( max_stack_count+1 );
					#endif
						return -1;
					}
				}
				else
				{
					// NbsO

					const float t0 = (xMin-poly.xMin)/oldInterval;

					if (t0 > 1.0e-4f)
					{
						for (int i=1; i<=degree; i++)
							for (int j=0; j<=degree-i; j++)
								poly.yCoords[j] += (poly.yCoords[j+1] - poly.yCoords[j]) * t0;
					}

					// ENbsO

					const float t1 = (poly.xMax-xMax)/(poly.xMax-xMin);

					if (t1 > 1.0e-4f)
					{
						for (int i=1; i<=degree; i++)
							for (int j=degree; j>=i; j--)
								poly.yCoords[j] += (poly.yCoords[j-1] - poly.yCoords[j]) * t1;
					}

					poly.xMin = xMin;
					poly.xMax = xMax;
				}
			}
			else	// nIntersections >= 2
			{
				CtrlPolygon newLeft;
				memcpy(newLeft.yCoords, poly.yCoords, sizeof(float)*(degree+1));
				
				for (int i=1; i<=degree; i++)
					for (int j=degree; j>=i; j--)
						newLeft.yCoords[j] += (newLeft.yCoords[j-1] - newLeft.yCoords[j]) * 0.5f;

				for (int i=1; i<=degree; i++)
					for (int j=0; j<=degree-i; j++)
						poly.yCoords[j] += (poly.yCoords[j+1] - poly.yCoords[j]) * 0.5f;

				// poly  newLeft  {xMin, xMax} ̍XV
				const float divX = poly.xMin + (poly.xMax - poly.xMin) * 0.5f;
				newLeft.xMin = poly.xMin;
				newLeft.xMax = divX;
				poly.xMin = divX;

				poly.nIntersections = countPolygonIntersections(poly.yCoords);
				if ( ! poly.nIntersections)
				{
					stack_index--;
				}

				newLeft.nIntersections = countPolygonIntersections(newLeft.yCoords);
				if ( newLeft.nIntersections )
				{
					// X^bN̏ɒBI
					if (++stack_index == polygon_stack_size)
					{
					#ifdef USE_COUNTER
						iterations_for_not_found.push_back( counter );
						max_stack_counts.push_back( max_stack_count+1 );
					#endif
						return -1;
					}

					// ̐_X^bNɐς
					// 0 ɋ߂_悤

					polygon_stack[stack_index] = newLeft;

				#ifdef USE_COUNTER
					max_stack_count = max(max_stack_count, stack_index);
				#endif
				}

				if (stack_index < 0)
				{
				#ifdef USE_COUNTER
					iterations_for_not_found.push_back( counter );
					max_stack_counts.push_back( max_stack_count+1 );
				#endif
					return -1;
				}
			}
		}

	#ifdef USE_COUNTER
		iterations_for_not_found.push_back( num_loop_limit );
		max_stack_counts.push_back( max_stack_count+1 );
	#endif

		return -1.f;
	}

private:
	struct CtrlPolygon
	{
		float yCoords[degree+1];
		float xMin;
		float xMax;
		int nIntersections;
	};

	float initial_coeffs[degree+1];

	inline int countPolygonIntersections(const float *yCoords) const
	{
		int nInts = 0;

		for (int i=0; i<degree; i++)
			nInts += ((yCoords[i] > 0) != (yCoords[i+1] > 0));

		return nInts;
	}

	// y W̕ς̂ 1 xo[W
	float findXMin(const CtrlPolygon &poly) const
	{
		const float *yCoords = poly.yCoords;

		int kIdx = -1;
		float maxAbsSlope = 0.f;

		for (int i=degree; i>=1; i--)
		{
			if (yCoords[i] * yCoords[0] > 0.f) // 
				break;

			const float absSlope = fabs(yCoords[i]-yCoords[0]) / (float)i;
			if (maxAbsSlope < absSlope)
			{
				maxAbsSlope = absSlope;
				kIdx = i;
			}
		}

		float minSlope = 10000.f;

		for (int i=0; i<kIdx; i++)
		{
			if (yCoords[i] * yCoords[kIdx] > 0.f) // 
				break;

			const float slope = (yCoords[kIdx]-yCoords[i]) / (float)(kIdx - i);
			minSlope = (fabs(slope) < fabs(minSlope)) ? slope : minSlope;
		}
		
		const float xCoeff = (poly.xMax - poly.xMin) / (float)degree;

		return poly.xMin + xCoeff * (kIdx - yCoords[kIdx]/minSlope);
	}

	// y W̕ς̂ 1 xo[W
	float findXMax(const CtrlPolygon &poly) const
	{
		const float *yCoords = poly.yCoords;

		int kIdx = -1;
		float maxAbsSlope = 0;

		for (int i=0; i<degree; i++)
		{
			if (yCoords[i] * yCoords[degree] > 0.f) // 
				break;

			const float absSlope = fabs(yCoords[degree]-yCoords[i])/(float)(degree-i);
			if (maxAbsSlope < absSlope)
			{
				maxAbsSlope = absSlope;
				kIdx = i;
			}
		}

		float minSlope = 10000.f;

		for (int i=degree; i>kIdx; i--)
		{
			if (yCoords[i] * yCoords[kIdx] > 0.f) // 
				break;

			const float slope = (yCoords[i]-yCoords[kIdx])/(float)(i-kIdx);
			minSlope = (fabs(slope) < fabs(minSlope)) ? slope : minSlope;
		}

		const float xCoeff = (poly.xMax - poly.xMin) / (float)degree;

		return poly.xMin + xCoeff * (kIdx - yCoords[kIdx]/minSlope);
	}

	//float Newton(float t) const
	//{
	//	return t - evaluate(t)/evaluateDifferential(t);
	//}

	//float evaluate(float t) const
	//{
	//	float tmp[degree+1];
	//	memcpy(tmp, initial_coeffs, sizeof(float)*(degree+1));

	//	for (int i=1; i<=degree; i++)
	//		for (int j=0; j<=degree-i; j++)
	//			tmp[j] += (tmp[j+1] - tmp[j]) * t;

	//	return tmp[0];
	//}

	//float evaluateDifferential(float t) const
	//{
	//	float tmp[degree+1];

	//	for (int i=0; i<degree; i++)
	//		tmp[i] = -initial_coeffs[i] + initial_coeffs[i+1];

	//	tmp[degree] = 0.f;

	//	for (int i=2; i<=degree; i++)
	//		for (int j=0; j<=degree-i; j++)
	//			tmp[j] += (tmp[j+1] - tmp[j]) * t;

	//	return degree*tmp[0];
	//}

	void polynomial2BezierForm(const float* polyCoeffs, float *coeffs)
	{
		memcpy(coeffs, polyCoeffs, sizeof(float)*(degree+1));

		for (int i=1; i<=degree; i++)
		{
			const float c = 1.f/(degree + 1.f - i);
			float d = 1.f;
			float e = c;
			for (int j=degree; j>=i; j--)
			{
				coeffs[j] = d * coeffs[j] + e * coeffs[j-1];
				d -= c;
				e += c;
			}
		}
	}
};

