#pragma once

#include <vector>
#include "GeometricObject.h"
#include "LightSource.h"
#include "EnvironmentMap.h"

class Scene
{
public:
	Scene()
		: m_pEnvironmentMap(0), m_BackgroundColor(1,1,1)
	{
	}

	~Scene()
	{
		for (int i=0; i<(int)m_Objects.size(); i++)
			delete m_Objects[i];
		
		for (int i=0; i<(int)m_Lights.size(); i++)
			delete m_Lights[i];
	}

	// objects

	int getNumObjects() const { return (int)m_Objects.size(); }

	const std::vector<GeometricObject *> &getObjects() const { return m_Objects; } 
	std::vector<GeometricObject *> &getObjects() { return m_Objects; } 

	GeometricObject *getObject(int i) { return m_Objects[i]; }
	void addObject(GeometricObject *o)
	{
		m_Objects.push_back( o );

		MyAlgebra::vec3 c(frand(), frand(), frand());
		c /= c.r + c.g + c.b;

		m_PseudoColors.push_back( c );
	}

	const MyAlgebra::vec3 &getPseudoColor(int i) const { return m_PseudoColors[i]; }

	// light sources

	int getNumLights() const { return (int)m_Lights.size(); }

	const std::vector<LightSource *> &getLights() const { return m_Lights; }
	std::vector<LightSource *> &getLights() { return m_Lights; }

	LightSource *getLight(int i) { return m_Lights[i]; }
	void addLight(LightSource *l) { m_Lights.push_back( l ); }

	bool isEnvironmentMapAvailable() const { return (m_pEnvironmentMap != 0); }
	EnvironmentMap *getEnvironmentMap() { return m_pEnvironmentMap; }
	void setEnvironmentMap(EnvironmentMap *pMap)
	{
		if ( ! m_pEnvironmentMap) delete m_pEnvironmentMap;
		m_pEnvironmentMap = pMap;
	} 

	MyAlgebra::vec3 getBackgroundColor() const { return m_BackgroundColor; }
	void setBackgroundColor(const MyAlgebra::vec3 &c) { m_BackgroundColor = c; }

	void clearScene()
	{
		//for (int i=0; i<(int)m_Objects.size(); i++)
		//	delete m_Objects[i];
		m_Objects.clear();

		m_PseudoColors.clear();

		for (int i=0; i<(int)m_Lights.size(); i++)
			delete m_Lights[i];
		m_Lights.clear();

		if (m_pEnvironmentMap)
		{
			delete m_pEnvironmentMap;
			m_pEnvironmentMap = 0;
		}
	}

private:
	std::vector<LightSource *> m_Lights;
	std::vector<GeometricObject *> m_Objects;
	std::vector<MyAlgebra::vec3> m_PseudoColors;

	EnvironmentMap *m_pEnvironmentMap;

	MyAlgebra::vec3 m_BackgroundColor;

	// returns value in [-1, 1]
	inline float frand() { return 2.f * (((float)rand()/(float)RAND_MAX) - 0.5f); }
};

