#pragma once

#include <sstream>
#include <string>
#include "Camera.h"
#include "Scene.h"
#include "Material.h"
#include "Sampler.h"
#include "Filter.h"

class EnvironmentMap;

class SceneLoader
{
public:
	SceneLoader()
		: error_in_line(false), line_number(0), current_material(0)
	{
	}

	~SceneLoader()
	{
		for (int i=0; i<(int)lights.size(); i++)
			delete lights[i];

		for (int i=0; i<(int)objects.size(); i++)
			delete objects[i];
	}

	bool load(Scene &scene, const char *filename);

private:
	bool error_in_line;
	std::istringstream iss;
	int line_number;

	Material *current_material;
	EnvironmentMap *specified_envmap;
	Sampler *pixel_sampler;
	Filter *pixel_filter;

	int screen_width, screen_height;
	MyAlgebra::vec3 camera_position;
	MyAlgebra::vec3 camera_center;
	MyAlgebra::vec3 camera_up;
	MyAlgebra::Real camera_fovy;
	int _max_recursion_depth;
	MyAlgebra::vec3 _background_color;
	float direct_light_scaling;
	std::vector<LightSource *> lights;
	std::vector<GeometricObject *> objects;

	bool parseLine();
	bool requireChar(char c);
	bool requireEol();
	bool readFloat(float &f);
	bool readInteger(int &i);
	bool readVector(MyAlgebra::vec3 &v);
	bool readString(std::string &str);

	bool checkEol();
	void checkCurrentMaterial();
	void reportWarning(const char *str);
	void reportError(const char *str);
};
