#pragma once

#include "Texture.h"

class SolidCheckeredTexture : public Texture
{
private:
	SolidCheckeredTexture()
		: u_num_repeats(3), v_num_repeats(3), w_num_repeats(3)
	{
		u_repeat_interval = 1.f / (float)u_num_repeats;
		v_repeat_interval = 1.f / (float)v_num_repeats;
		w_repeat_interval = 1.f / (float)w_num_repeats;
		color0.set(0.f);	// black
		color1.set(1.f);	// white
	}

public:
	static SolidCheckeredTexture *CreateTexture()
	{
		SolidCheckeredTexture *t = new SolidCheckeredTexture();
		s_TextureCache.push_back( t );
		return t;
	}

	~SolidCheckeredTexture()
	{
	}

	MyAlgebra::vec3 getColor(const MyAlgebra::vec3 &uvw) const
	{
		const float _u = fmodf(uvw.x, u_repeat_interval) / u_repeat_interval;
		const float _v = fmodf(uvw.y, v_repeat_interval) / v_repeat_interval;
		const float _w = fmodf(uvw.z, w_repeat_interval) / w_repeat_interval;

		return ((_u<=0.5 && _v<=0.5 && _w<=0.5) || (_u>0.5 && _v>0.5 && _w<=0.5) ||
				(_u<=0.5 && _v>0.5 && _w>0.5) || (_u>0.5 && _v<=0.5 && _w>0.5))
				? color0 : color1;
	}

	void setColors(const MyAlgebra::vec3 &c1, const MyAlgebra::vec3 &c2)
	{
		color0.set( c1 );
		color1.set( c2 );
	}

	void setColor0(const MyAlgebra::vec3 &c) { color0.set(c); }
	void setColor1(const MyAlgebra::vec3 &c) { color1.set(c); }
	void setColor0(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color0.set(r, g, b); }
	void setColor1(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color1.set(r, g, b); }

	MyAlgebra::vec3 getColor0() const { return color0; }
	MyAlgebra::vec3 getColor1() const { return color1; }

	void setUNumRepeats(int n)
	{
		u_num_repeats = n;
		u_repeat_interval = 1.f / (float)u_num_repeats;
	}

	void setVNumRepeats(int n)
	{
		v_num_repeats = n;
		v_repeat_interval = 1.f / (float)v_num_repeats;
	}

	void setWNumRepeats(int n)
	{
		w_num_repeats = n;
		w_repeat_interval = 1.f / (float)w_num_repeats;
	}

	int getUNumRepeats() const { return u_num_repeats; }
	int getVNumRepeats() const { return v_num_repeats; }
	int getWNumRepeats() const { return w_num_repeats; }

	Texture_Type getTextureType() const { return Solid_Checker_Texture; }
	bool isSolidTexture() const { return true; }

private:
	int u_num_repeats;	// number of repeating the texture pattern within [0, 1]
	int v_num_repeats;	// number of repeating the texture pattern within [0, 1]
	int w_num_repeats;	// number of repeating the texture pattern within [0, 1]
	float u_repeat_interval;	// interval for repeating the texture pattern
	float v_repeat_interval;	// interval for repeating the texture pattern
	float w_repeat_interval;	// interval for repeating the texture pattern
	MyAlgebra::vec3 color0;
	MyAlgebra::vec3 color1;

};
