#include "SolidNoise.h"
#include "RNG.h"

// from "Realistic Ray Tracing, second ed.", p. 85

using namespace MyAlgebra;

SolidNoise::SolidNoise()
{
	RNG random;

	grad[0] = vec3( 1, 1, 0);
	grad[1] = vec3(-1, 1, 0);
	grad[2] = vec3( 1,-1, 0);
	grad[3] = vec3(-1,-1, 0);

	grad[4] = vec3( 1, 0, 1);
	grad[5] = vec3(-1, 0, 1);
	grad[6] = vec3( 1, 0,-1);
	grad[7] = vec3(-1, 0,-1);

	grad[8] =  vec3( 0, 1, 1);
	grad[9] =  vec3( 0,-1, 1);
	grad[10] = vec3( 0, 1,-1);
	grad[11] = vec3( 0,-1,-1);

	grad[12] = vec3( 1, 1, 0);
	grad[13] = vec3(-1, 1, 0);
	grad[14] = vec3( 0,-1, 1);
	grad[15] = vec3( 0,-1,-1);

	for (int i=0; i<num_vectors; i++)
		phi[i] = i;

	// shuffle phi
	for (int i=num_vectors-2; i>=0; i--)
	{
		const int target = int(random()*i);
		const int temp = phi[i+1];
		phi[i+1] = phi[target];
		phi[target] = temp;
	}
}

float SolidNoise::turbulence(const vec3 &p, int depth) const
{
	float weight = 1.f;
	vec3 ptemp( p );

	float sum = fabs(noise(ptemp));

	for (int i=1; i<depth; i++)
	{
		weight *= 2.f;
		ptemp = weight * p;

		sum += fabs(noise(ptemp)) / weight;
	}

	return sum;
}

float SolidNoise::dturbulence(const vec3 &p, int depth, float d) const
{
	float weight = 1.f;
	vec3 ptemp( p );

	float sum = fabs(noise(ptemp));

	for (int i=1; i<depth; i++)
	{
		weight *= d;
		ptemp = weight * p;

		sum += fabs(noise(ptemp)) / d;
	}

	return sum;
}

float SolidNoise::noise(const vec3 &p) const
{
	const int fi = int(floor(p.x));
	const int fj = int(floor(p.y));
	const int fk = int(floor(p.z));

	const float fu = p.x - fi;
	const float fv = p.y - fj;
	const float fw = p.z - fk;

	float sum = 0.f;

	sum += knot(fi, fj, fk,			vec3(fu, fv, fw));

	sum += knot(fi+1, fj, fk,		vec3(fu-1, fv, fw));

	sum += knot(fi, fj+1, fk,		vec3(fu, fv-1, fw));

	sum += knot(fi, fj, fk+1,		vec3(fu, fv, fw-1));

	sum += knot(fi+1, fj+1, fk,		vec3(fu-1, fv-1, fw));

	sum += knot(fi+1, fj, fk+1,		vec3(fu-1, fv, fw-1));

	sum += knot(fi, fj+1, fk+1,		vec3(fu, fv-1, fw-1));

	sum += knot(fi+1, fj+1, fk+1,	vec3(fu-1, fv-1, fw-1));

	return sum;
}