#pragma once

#include "my_algebra.h"

// from "Realistic Ray Tracing, second ed.", p. 85

class SolidNoise
{
public:
	SolidNoise();

	~SolidNoise()
	{
	}

	float turbulence(const MyAlgebra::vec3 &p, int depth) const;
	float dturbulence(const MyAlgebra::vec3 &p, int depth, float d) const;
	float noise(const MyAlgebra::vec3 &p) const;

private:
	static const int num_vectors = 16;
	MyAlgebra::vec3 grad[num_vectors];
	int phi[num_vectors];

	inline float omega(float t) const
	{
		// assume that t is within [-1, 1]
		if (t < 0.f) t = -t;
		const float t2 = t*t;
		return 1.f + t*t2 * (-10.f + t * (15.f - 6.f * t));
	}

	inline MyAlgebra::vec3 gamma(int i, int j, int k) const
	{
		int idx = phi[ abs(k) % num_vectors ];
		idx = phi[abs(j + idx) % num_vectors];
		idx = phi[abs(i + idx) % num_vectors];
		return grad[idx];
	}

	inline MyAlgebra::Real knot(int i, int j, int k, MyAlgebra::vec3 &v) const
	{
		return omega(v.x) * omega(v.y) * omega(v.z) * dot(gamma(i,j,k),v);
	}

	inline int intGamma(int i, int j) const
	{
		int idx = phi[ abs(j) % num_vectors ];
		idx = phi[ abs(i + idx) % num_vectors ];
		return idx;
	}
};
