#include "SolidTexturedSphere.h"
#include <cstdlib>
#include <cmath>
#include <GL/glut.h>

using namespace std;
using namespace MyAlgebra;

bool SolidTexturedSphere::hit(const Ray &r, Real tmin, Real tmax, HitRecord &record) const
{
	const vec3 v = r.getOrigin() - center;
	const vec3 d = r.getUnitDir();

	const Real b = dot(v, d);
	const Real c = v.sq_norm() - radius*radius;

	const Real D = b*b - c;

	if (D > 0.0)
	{
		const Real sqrtD = sqrtf(D);
		Real t = -b - sqrtD;

		// check if t is within a valid interval
		if (t < tmin)
			t = -b + sqrtD;
		if (t < tmin || t > tmax)
			return false;

		// valid hit

		vec3 n = r.getPosition( t ) - center;
		n.normalize();

		record.t = t;
		record.normal = n;
		record.hit_pos = r.getPosition(t);

		// TODO: calculate uvw texture coordinates that are within [0, 1]^3
		// The texture coordinates will cover over the bounding cube of this sphere

		// record.tex_coords = ??

		record.material = material;

		return true;
	}

	return false;
}

bool SolidTexturedSphere::shadowHit(const Ray &r, Real tmin, Real tmax) const
{
	const vec3 v = r.getOrigin() - center;
	const vec3 d = r.getUnitDir();

	const Real b = dot(v, d);
	const Real c = v.sq_norm() - radius*radius;

	const Real D = b*b - c;

	if (D > 0.0)
	{
		const Real sqrtD = sqrtf(D);
		Real t = -b - sqrtD;

		// check if t is within a valid interval
		if (t < tmin)
			t = -b + sqrtD;
		if (t < tmin || t > tmax)
			return false;

		// valid hit

		return true;
	}

	return false;
}

// for preview using OpenGL
void SolidTexturedSphere::drawGL() const
{
	const int nSplit = 16;

	glPushMatrix();
	glTranslatef( center.x, center.y, center.z );
	glutSolidSphere( radius, nSplit, nSplit );
	glPopMatrix();
}
