#pragma once

#include "GeometricObject.h"

class Sphere : public GeometricObject
{
private:
	Sphere()
	{
	}

	Sphere(const Sphere &s)
		: center(s.center), radius(s.radius)
	{
		material = s.material;
	}

	Sphere(MyAlgebra::vec3 &_center, MyAlgebra::Real r, Material *m)
		: center(_center), radius(r)
	{
		material = m;
	}

public:
	static Sphere *CreateGeometricObject()
	{
		Sphere *obj = new Sphere();
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static Sphere *CreateGeometricObject(const Sphere &s)
	{
		Sphere *obj = new Sphere(s);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static Sphere *CreateGeometricObject(MyAlgebra::vec3 &_center, MyAlgebra::Real r, Material *m)
	{
		Sphere *obj = new Sphere(_center, r, m);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	bool hit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const;
	bool shadowHit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const;

	void drawGL() const;	// for preview using OpenGL

	inline MyAlgebra::vec3 getCenter() const { return center; }
	inline MyAlgebra::Real getRadius() const { return radius; }

	inline void setCenter(const MyAlgebra::vec3 &c) { center = c; }
	inline void setRadius(MyAlgebra::Real r) { radius = r; }

private:
	MyAlgebra::vec3 center;
	MyAlgebra::Real radius;

};

