#include "TexturedPlane.h"
#include "TexturedMaterial.h"
#include <cstdlib>
#include <GL/glut.h>

using namespace std;
using namespace MyAlgebra;

bool TexturedPlane::hit(const Ray &r, Real tmin, Real tmax, HitRecord &record) const
{
	if ( ! material || material->getMaterialType() != Material::Textured_Type)
	{
		cerr << __FUNCTION__ << ": invalid material" << endl;
		return false;
	}

	const Texture *tex = ((TexturedMaterial *)material)->getTexture();

	if ( ! tex)
	{
		cerr << __FUNCTION__ << ": texture not set" << endl;
		return false;
	}

	const vec3 dir = r.getUnitDir();

	const Real a = dot(dir, normal);

	if (fabs(a) < 1.0e-10) return false;

	const Real b = dot(r.getOrigin(), normal);

	const Real t = -(b - d) / a;

	if (t < tmin || t > tmax) return false;

	// calculate texture coordinates
	const vec3 hitPos = r.getPosition( t );

	record.t = t;
	record.normal = normal;
	record.hit_pos = hitPos;

	// TODO: calculate uv texture coordinates that are within [0, 1]^2
	// First, project the origin onto this plane in order to use as a reference point
	// Then, calculate the local coordinates using local axes, 
	// i.e., binormal and tangent (member variables in this class)
	// Finally, calculate the modulo of the local coordinates using texture_scale
	// (member variable) in order to get values within [0, 1]^2

	// record.tex_coords.x = ??
	// record.tex_coords.y = ??

	record.material = material;

	return true;
}

bool TexturedPlane::shadowHit(const Ray &r, Real tmin, Real tmax) const
{
	const vec3 dir = r.getUnitDir();

	const Real a = dot(dir, normal);

	if (fabs(a) < 1.0e-10) return false;

	const Real b = dot(r.getOrigin(), normal);

	const Real t = -(b - d) / a;

	if (t < tmin || t > tmax) return false;

	return true;
}

// for preview using OpenGL
void TexturedPlane::drawGL() const
{
	vec3 u,v;
	normal.getTwoOrthogonals(u, v);

	const Real s = 1.0e+5;
	u.scale( s );
	v.scale( s );

	const vec3 p = d * normal;

	glBegin(GL_QUADS);
	glVertex3fv( p - u - v );
	glVertex3fv( p + u - v );
	glVertex3fv( p + u + v );
	glVertex3fv( p - u + v );
	glEnd();
}
