#pragma once

#include "GeometricObject.h"

class TexturedPlane : public GeometricObject
{
private:
	TexturedPlane()
		: texture_scale(10.f)
	{
	}

	TexturedPlane(const TexturedPlane &p)
		: normal(p.normal), binormal(p.binormal), tangent(p.tangent), d(p.d), texture_scale(p.texture_scale)
	{
		//p.material->clone(&material);
		material = p.material;
	}

	TexturedPlane(const MyAlgebra::vec3 &n, MyAlgebra::Real _d, Material *m)
		: d(_d), texture_scale(10.f)
	{
		material = m;
		setNormal(n);
	}

public:
	static TexturedPlane *CreateGeometricObject()
	{
		TexturedPlane *obj = new TexturedPlane();
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static TexturedPlane *CreateGeometricObject(const TexturedPlane &p)
	{
		TexturedPlane *obj = new TexturedPlane(p);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static TexturedPlane *CreateGeometricObject(const MyAlgebra::vec3 &n, MyAlgebra::Real _d, Material *m)
	{
		TexturedPlane *obj = new TexturedPlane(n, _d, m);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	bool hit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const;
	bool shadowHit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const;

	void drawGL() const;	// for preview using OpenGL

	inline MyAlgebra::vec3 getNormal() const { return normal; }
	inline MyAlgebra::Real getD() const { return d; }

	inline void setNormal(const MyAlgebra::vec3 &n)
	{
		normal = n; 
		normal.getTwoOrthogonals(binormal, tangent);
		binormal.normalize();
		tangent.normalize();
	}

	inline void setD(MyAlgebra::Real _d) { d = _d; }

	inline void setTextureScale(MyAlgebra::Real s) { texture_scale = s; }
	inline MyAlgebra::Real getTextureScale() const { return texture_scale; }	

private:
	MyAlgebra::vec3 normal;
	MyAlgebra::vec3 binormal;	// axis for local coordinate, used for texture mapping
	MyAlgebra::vec3 tangent;	// axis for local coordinate, used for texture mapping
	MyAlgebra::Real texture_scale;	// length for repeating the texture pattern
	MyAlgebra::Real d;
};

