#include "TexturedSphere.h"
#include <cstdlib>
#include <cmath>
#include <GL/glut.h>

using namespace std;
using namespace MyAlgebra;

bool TexturedSphere::hit(const Ray &r, Real tmin, Real tmax, HitRecord &record) const
{
	const vec3 v = r.getOrigin() - center;
	const vec3 d = r.getUnitDir();

	const Real b = dot(v, d);
	const Real c = v.sq_norm() - radius*radius;

	const Real D = b*b - c;

	if (D > 0.0)
	{
		const Real sqrtD = sqrtf(D);
		Real t = -b - sqrtD;

		// check if t is within a valid interval
		if (t < tmin)
			t = -b + sqrtD;
		if (t < tmin || t > tmax)
			return false;

		// valid hit

		vec3 n = r.getPosition( t ) - center;
		n.normalize();

		record.t = t;
		record.normal = n;
		record.hit_pos = r.getPosition(t);

		// TODO: calculate uv texture coordinates that are within [0, 1]^2
		// Calculate the longitude/latitude of the sphere, and normalize them 
		// to obtain the texture coordinates
		// The y axis plays the role of the polar axis

		// record.tex_coords.x = ??
		// record.tex_coords.y = ??

		record.material = material;

		return true;
	}

	return false;
}

bool TexturedSphere::shadowHit(const Ray &r, Real tmin, Real tmax) const
{
	const vec3 v = r.getOrigin() - center;
	const vec3 d = r.getUnitDir();

	const Real b = dot(v, d);
	const Real c = v.sq_norm() - radius*radius;

	const Real D = b*b - c;

	if (D > 0.0)
	{
		const Real sqrtD = sqrtf(D);
		Real t = -b - sqrtD;

		// check if t is within a valid interval
		if (t < tmin)
			t = -b + sqrtD;
		if (t < tmin || t > tmax)
			return false;

		// valid hit

		return true;
	}

	return false;
}

// for preview using OpenGL
void TexturedSphere::drawGL() const
{
	glPushMatrix();
	glTranslatef( center.x, center.y, center.z );
	glutSolidSphere( radius, 32, 32 );
	glPopMatrix();
}
