#pragma once

#include "GeometricObject.h"

class TexturedSphere : public GeometricObject
{
private:
	TexturedSphere()
	{
	}

	TexturedSphere(const TexturedSphere &s)
		: center(s.center), radius(s.radius)
	{
		//s.material->clone(&material);
		material = s.material;
	}

	TexturedSphere(MyAlgebra::vec3 &_center, MyAlgebra::Real r, Material *m)
		: center(_center), radius(r)
	{
		material = m;
	}

public:
	static TexturedSphere *CreateGeometricObject()
	{
		TexturedSphere *obj = new TexturedSphere();
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static TexturedSphere *CreateGeometricObject(const TexturedSphere &s)
	{
		TexturedSphere *obj = new TexturedSphere(s);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static TexturedSphere *CreateGeometricObject(MyAlgebra::vec3 &_center, MyAlgebra::Real r, Material *m)
	{
		TexturedSphere *obj = new TexturedSphere(_center, r, m);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	bool hit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const;
	bool shadowHit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const;

	void drawGL() const;	// for preview using OpenGL

	inline MyAlgebra::vec3 getCenter() const { return center; }
	inline MyAlgebra::Real getRadius() const { return radius; }

	inline void setCenter(const MyAlgebra::vec3 &c) { center = c; }
	inline void setRadius(MyAlgebra::Real r) { radius = r; }

private:
	MyAlgebra::vec3 center;
	MyAlgebra::Real radius;

};

