#include "Tokenizer.h"

#include <sstream>

using namespace std;
using namespace MyAlgebra;

bool Tokenizer::readToken( std::string &str )
{
	if ( skipWhiteSpaces() ) return true;	// G[ 

	char c = m_pIFS->get();

	//cerr << __FUNCTION__ << ": read char: " << c << endl;

	// {͂łǂނ̃g[Nǂ񂾂킩Ă̂
	// g[N̎ނ܂߂ĕԂĂׂ
	if ( ';' == c || '{' == c || '}' == c || '(' == c || ')' == c )
	{
		str = c;
		return false;
	}

	ostringstream oss;

	while ( isalnum(c) || '_' == c || '-' == c || '/' == c || ':' == c || '#' == c)
	{
		if (m_pIFS->eof()) break;
		oss << c;
		c = m_pIFS->get();
	}

	m_pIFS->unget();

	str = oss.str();

	//cerr << __FUNCTION__ << ": read string: " << str << endl;

	if (m_pIFS->fail())
	{
		cerr << __FUNCTION__ << ": cannot read token at line " << m_CurrentLineNumber << endl;
		m_ErrorFlag = true;
		return true; // G[ 
	}

	return false;	// G[Ȃ 
}

bool Tokenizer::requireChar( char c )
{
	if ( skipWhiteSpaces() ) return true;	// G[ 

	const char r = m_pIFS->get();

	if (m_pIFS->fail())
	{
		cerr << __FUNCTION__ << ": cannot read char at line " << m_CurrentLineNumber << endl;
		m_ErrorFlag = true;
		return true; // G[ 
	}
	else if (r != c)
	{
		cerr << __FUNCTION__ << ": '" << c << "' expected, but '" << r << "' read at line " << m_CurrentLineNumber << endl;
		m_ErrorFlag = true;
		return true; // G[ 
	}

	return false;	// G[Ȃ 
}

bool Tokenizer::requireString( const string &str )
{
	if ( skipWhiteSpaces() ) return true;	// G[ 

	ostringstream oss;

	char c = m_pIFS->get();
	while ( isalnum(c) || '_' == c || '-' == c || '/' == c || ':' == c || '#' == c || '.' == c )	// ̒`͗v
	{
		if (m_pIFS->eof()) break;
		oss << c;
		c = m_pIFS->get();
	}

	m_pIFS->unget();

	const string tmp = oss.str();

	if (m_pIFS->fail())
	{
		cerr << __FUNCTION__ << ": cannot read string at line " << m_CurrentLineNumber << endl;
		m_ErrorFlag = true;
		return true; // G[ 
	}
	else if (str != tmp)
	{
		cerr << __FUNCTION__ << ": \"" << str << "\" expected, but \"" << tmp << "\" read at line" << m_CurrentLineNumber << endl;
		m_ErrorFlag = true;
		return true; // G[ 
	}

	return false;	// G[Ȃ 
}


bool Tokenizer::readBoolean( bool &b )
{
	if ( skipWhiteSpaces() ) return true;	// G[ 

	string booleanExpression;
	m_ErrorFlag |= readString(booleanExpression);

	b = ( "true" == booleanExpression);

	if (m_pIFS->fail())
	{
		cerr << __FUNCTION__ << ": cannot read boolean at line " << m_CurrentLineNumber << endl;
		m_ErrorFlag = true;
		return true; // G[ 
	}

	return false;
}

bool Tokenizer::readFloat( float &f )
{
	if ( skipWhiteSpaces() ) return true;	// G[ 

	(*m_pIFS) >> f;

	if (m_pIFS->fail())
	{
		cerr << __FUNCTION__ << ": cannot read float at line " << m_CurrentLineNumber << endl;
		m_ErrorFlag = true;
		return true; // G[ 
	}

	return false;
}

bool Tokenizer::readInteger( int &i )
{
	if ( skipWhiteSpaces() ) return true;	// G[ 

	(*m_pIFS) >> i;

	if (m_pIFS->fail())
	{
		cerr << __FUNCTION__ << ": cannot read integer at line " << m_CurrentLineNumber << endl;
		m_ErrorFlag = true;
		return true; // G[ 
	}

	return false;
}

bool Tokenizer::readVec2( vec2 &v )
{
	m_ErrorFlag |= requireChar( '(' );
	m_ErrorFlag |= readFloat( v.x );
	m_ErrorFlag |= requireChar( ',' );
	m_ErrorFlag |= readFloat( v.y );
	m_ErrorFlag |= requireChar( ')' );

	return m_ErrorFlag;
}

bool Tokenizer::readVec3( vec3 &v )
{
	m_ErrorFlag |= requireChar( '(' );
	m_ErrorFlag |= readFloat( v.x );
	m_ErrorFlag |= requireChar( ',' );
	m_ErrorFlag |= readFloat( v.y );
	m_ErrorFlag |= requireChar( ',' );
	m_ErrorFlag |= readFloat( v.z );
	m_ErrorFlag |= requireChar( ')' );

	return m_ErrorFlag;
}

bool Tokenizer::readVec4( vec4 &v )
{
	m_ErrorFlag |= requireChar( '(' );
	m_ErrorFlag |= readFloat( v.x );
	m_ErrorFlag |= requireChar( ',' );
	m_ErrorFlag |= readFloat( v.y );
	m_ErrorFlag |= requireChar( ',' );
	m_ErrorFlag |= readFloat( v.z );
	m_ErrorFlag |= requireChar( ',' );
	m_ErrorFlag |= readFloat( v.w );
	m_ErrorFlag |= requireChar( ')' );

	return m_ErrorFlag;
}

bool Tokenizer::readFloats( float *f, int n )
{
	for (int i=0; i<n; i++)
	{
		m_ErrorFlag |= readFloat( f[i] );
		if ( i < n-1 )
			m_ErrorFlag |= requireChar( ',' );
	}

	return m_ErrorFlag;
}

bool Tokenizer::readString( string &str )
{
	if ( skipWhiteSpaces() ) return true;	// G[ 

	ostringstream oss;

	char c = m_pIFS->get();
	while ( isalnum(c) || '_' == c || '-' == c || '/' == c || ':' == c || '#' == c || '.' == c )	// ̒`͗v
	{
		if (m_pIFS->eof()) break;
		oss << c;
		c = m_pIFS->get();
	}

	m_pIFS->unget();

	str = oss.str();

	if (m_pIFS->fail())
	{
		cerr << __FUNCTION__ << ": cannot read string at line " << m_CurrentLineNumber << endl;
		m_ErrorFlag = true;
		return true;
	}

	//cerr << __FUNCTION__ << ": read: \"" << str << "\"" << endl;

	return false;
}

bool Tokenizer::readQuotedString( string &str )
{
	m_ErrorFlag |= requireChar( '"' );
	m_ErrorFlag |= readString( str );
	m_ErrorFlag |= requireChar( '"' );

	return m_ErrorFlag;
}

// private functions -----------------------------------------------------------------------

bool Tokenizer::skipWhiteSpaces()
{
	if (m_ErrorFlag) return true;

	int c;

	while ( c = m_pIFS->get() )
	{
		//cerr << __FUNCTION__ << ": read char: " << (char)c << endl; 

		if (m_pIFS->eof())
			return false;	// G[Ȃ
		else if ( '\n' == c )
			++m_CurrentLineNumber;
		else if ( isspace(c) )
			continue;
		else if ( '/' == c )
		{
			if ( skipComments(c) )
			{
				m_ErrorFlag = true;
				return true;	// G[ 
			}
		}
		else
		{
			//cerr << __FUNCTION__ << ": read: " << (char)c << ", break" << endl;
			break;	// 󔒈ȊOǂݍ 
		}
	}

	m_pIFS->unget();	// ]vɓǂݍ񂾂̂Ŗ߂ 

	return false;
}

bool Tokenizer::skipComments(int c)
{
	//cerr << __FUNCTION__ << ": input: " << (char)c << endl;

	if ( '/' == c )	// Rgǂ
	{
		c = m_pIFS->get();

		//cerr << __FUNCTION__ << ": read: " << (char)c << endl;

		if ( '/' == c ) // C++ X^C̃RgsȂ̂Ŏcǂݎ̂ 
		{
			string dummy;
			getline(*m_pIFS, dummy);
			++m_CurrentLineNumber;
		}
		else if ( '*' == c ) // C X^C̃Rg 
		{
			const int startLineNo = m_CurrentLineNumber;

			//cerr << __FUNCTION__ << ": C-style comment started" << endl;

			while ( true )
			{
				c = m_pIFS->get();

				//cerr << __FUNCTION__ << ": read in while: " << (char)c << endl;

				if ( '*' == c)
				{
					//cerr << __FUNCTION__ << ": '*' read ... terminate?" << endl;

					c = m_pIFS->get();

					//cerr << __FUNCTION__ << ": read2 in while: " << (char)c << endl;

					if ('/' == c )
					{
						//cerr << __FUNCTION__ << ": comment terminated" << endl;
						break;
					}
					else if ( '\n' == c )
					{
						//cerr << __FUNCTION__ << ": new line" << endl;
						++m_CurrentLineNumber;
					}
				}
				else if ( '\n' == c )
				{
					//cerr << __FUNCTION__ << ": new line" << endl;
					++m_CurrentLineNumber;
				}
				else if ( m_pIFS->eof() )
				{
					cerr << __FUNCTION__ << ": unterminated C-style comment from line " << startLineNo << endl;
					return true;	// G[ 
				}
			}
		}
		else
		{
			cerr << __FUNCTION__ << ": unexpected character '" << (char)c << "' at line " << m_CurrentLineNumber << endl;
			return true;	// G[ 
		}
	}

	return false;	// G[Ȃ 
}



