#pragma once

#include "GeometricObject.h"

class Triangle : public GeometricObject
{
public:
	static Triangle *CreateGeometricObject()
	{
		Triangle *obj = new Triangle();
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static Triangle *CreateGeometricObject(const Triangle &tri)
	{
		Triangle *obj = new Triangle(tri);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	static Triangle *CreateGeometricObject(const MyAlgebra::vec3 &_v0, const MyAlgebra::vec3 &_v1, const MyAlgebra::vec3 &_v2, Material *m)
	{
		Triangle *obj = new Triangle(_v0, _v1, _v2, m);
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	// TriangleMesh NX vector Ɋi[̂ŁARXgN^ public ɂKv 

	Triangle()
	{
		m_Texcoords[0].set( 0.f );
		m_Texcoords[1].set( 0.f );
	}

	Triangle(const Triangle &tri)
	{
		m_Vertices[0] = tri.m_Vertices[0];
		m_Vertices[1] = tri.m_Vertices[1];
		m_Vertices[2] = tri.m_Vertices[2];

		m_Normals[0] = tri.m_Normals[0];
		m_Normals[1] = tri.m_Normals[1];
		m_Normals[2] = tri.m_Normals[2];

		m_Texcoords[0] = tri.m_Texcoords[0];
		m_Texcoords[1] = tri.m_Texcoords[1];

		material = tri.material;
	}

	Triangle(const MyAlgebra::vec3 &_v0, const MyAlgebra::vec3 &_v1, const MyAlgebra::vec3 &_v2, Material *m)
	{
		m_Vertices[0] = _v0;
		m_Vertices[1] = _v1;
		m_Vertices[2] = _v2;

		m_Normals[0] = m_Normals[1] = m_Normals[2] = getNormal();

		m_Texcoords[0].set( 0.f );
		m_Texcoords[1].set( 0.f );

		material = m;
	}

	bool hit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const;
	bool shadowHit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const;

	void drawGL() const;	// for preview using OpenGL

	MyAlgebra::vec3 getNormal() const
	{ 
		MyAlgebra::vec3 n = (m_Vertices[1] - m_Vertices[0]) ^ (m_Vertices[2] - m_Vertices[0]);
		n.normalize();
		return n; 
	}

	inline MyAlgebra::vec3 getVertex0() const { return m_Vertices[0]; }
	inline MyAlgebra::vec3 getVertex1() const { return m_Vertices[1]; }
	inline MyAlgebra::vec3 getVertex2() const { return m_Vertices[2]; }

	inline void setVertex0(const MyAlgebra::vec3 &_v0) { m_Vertices[0] = _v0; }
	inline void setVertex1(const MyAlgebra::vec3 &_v1) { m_Vertices[1] = _v1; }
	inline void setVertex2(const MyAlgebra::vec3 &_v2) { m_Vertices[2] = _v2; }

	inline MyAlgebra::vec3 getNormal0() const { return m_Normals[0]; }
	inline MyAlgebra::vec3 getNormal1() const { return m_Normals[1]; }
	inline MyAlgebra::vec3 getNormal2() const { return m_Normals[2]; }

	inline void setNormal0(const MyAlgebra::vec3 &_n0) { m_Normals[0] = _n0; }
	inline void setNormal1(const MyAlgebra::vec3 &_n1) { m_Normals[1] = _n1; }
	inline void setNormal2(const MyAlgebra::vec3 &_n2) { m_Normals[2] = _n2; }

	inline MyAlgebra::vec2 getTexcoord0() const { return m_Texcoords[0]; }
	inline MyAlgebra::vec2 getTexcoord1() const { return m_Texcoords[1]; }
	inline MyAlgebra::vec2 getTexcoord2() const { return m_Texcoords[2]; }

	inline void setTexcoord0(const MyAlgebra::vec2 &_t0) { m_Texcoords[0] = _t0; }
	inline void setTexcoord1(const MyAlgebra::vec2 &_t1) { m_Texcoords[1] = _t1; }
	inline void setTexcoord2(const MyAlgebra::vec2 &_t2) { m_Texcoords[2] = _t2; }

	MyAlgebra::vec3 getBoundingBoxMin() const
	{
		const float x = std::min( m_Vertices[0].x, std::min(m_Vertices[1].x, m_Vertices[2].x));
		const float y = std::min( m_Vertices[0].y, std::min(m_Vertices[1].y, m_Vertices[2].y));
		const float z = std::min( m_Vertices[0].z, std::min(m_Vertices[1].z, m_Vertices[2].z));
		return MyAlgebra::vec3(x,y,z);
	}
	
	MyAlgebra::vec3 getBoundingBoxMax() const
	{
		const float x = std::max( m_Vertices[0].x, std::max(m_Vertices[1].x, m_Vertices[2].x));
		const float y = std::max( m_Vertices[0].y, std::max(m_Vertices[1].y, m_Vertices[2].y));
		const float z = std::max( m_Vertices[0].z, std::max(m_Vertices[1].z, m_Vertices[2].z));
		return MyAlgebra::vec3(x,y,z);
	}

private:
	MyAlgebra::vec3 m_Vertices[3];
	MyAlgebra::vec3 m_Normals[3];
	MyAlgebra::vec2 m_Texcoords[3];

};

