#pragma once

#include "Triangle.h"
#include <vector>
#include "UniformGrid.h"

class TriangleMeshUniformGrid : public GeometricObject
{
private:
	TriangleMeshUniformGrid() : m_ShadingType(Flat_Shading)
	{
	}

public:
	enum Shading_Type
	{
		Flat_Shading,
		Smooth_Shading,	// Phong smooth shading
	};

	static TriangleMeshUniformGrid *CreateGeometricObject()
	{
		TriangleMeshUniformGrid *obj = new TriangleMeshUniformGrid();
		s_GeometricObjectCache.push_back( obj );
		return obj;
	}

	// class definitions

	bool hit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax, HitRecord &record) const;
	bool shadowHit(const Ray &r, MyAlgebra::Real tmin, MyAlgebra::Real tmax) const;

	void drawGL() const;	// for preview using OpenGL

	void clear() { m_Triangles.clear(); }

	int getNumTriangles() const { return m_Triangles.size(); }

	// setter/getter

	void addTriangle(const Triangle &t) { m_Triangles.push_back(t); }
	void setTriangle(int i, const Triangle &t) { m_Triangles[i] = t; }

	const std::vector<Triangle> &getTriangles() const { return m_Triangles; }
	std::vector<Triangle> &getTriangles() { return m_Triangles; }

	const Triangle &getTriangle(int i) const { return m_Triangles[i]; }
	Triangle &getTriangle(int i) { return m_Triangles[i]; }

	void computeBoundingBox();

	void setBoundingBoxMin(const MyAlgebra::vec3 &p) { m_BoundingBoxPos[0].set( p ); }
	void setBoundingBoxMax(const MyAlgebra::vec3 &p) { m_BoundingBoxPos[1].set( p ); }

	MyAlgebra::vec3 getBoundingBoxMin() const { return m_BoundingBoxPos[0]; }
	MyAlgebra::vec3 getBoundingBoxMax() const { return m_BoundingBoxPos[1]; }

	void setShadingType(Shading_Type type) { m_ShadingType = type; }
	Shading_Type getShadingType() const { return m_ShadingType; }

	void buidGrid(int nXCells, int nYCells, int nZCells);

private:
	Shading_Type m_ShadingType;

	std::vector<Triangle> m_Triangles;

	MyAlgebra::vec3 m_BoundingBoxPos[2];

	UniformGrid m_Grid;
};

