#pragma once

#include "Texture.h"
#include "SolidNoise.h"

class TurbulenceTexture : public Texture
{
private:
	TurbulenceTexture()
		: scale(1.f), octaves(5),
		color0(0.8f), color1(0.f, 0.f, 1.f)
	{
	}

public:
	static TurbulenceTexture *CreateTexture()
	{
		TurbulenceTexture *t = new TurbulenceTexture();
		s_TextureCache.push_back( t );
		return t;
	}

	MyAlgebra::vec3 getColor(const MyAlgebra::vec3 &uvw) const
	{
		const MyAlgebra::Real t = (1.f + solid_noise.turbulence(scale*uvw, octaves)) / 2.f;
		return t * color0 + (1.f-t) * color1;
	}

	float getScale() const { return scale; }
	void setScale(float s) { scale = s; }

	void setColor0(const MyAlgebra::vec3 &c) { color0.set(c); }
	void setColor1(const MyAlgebra::vec3 &c) { color1.set(c); }
	void setColor0(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color0.set(r, g, b); }
	void setColor1(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color1.set(r, g, b); }

	MyAlgebra::vec3 getColor0() const { return color0; }
	MyAlgebra::vec3 getColor1() const { return color1; }

	Texture_Type getTextureType() const { return Turbulence_Texture; }
	bool isSolidTexture() const { return true; }

private:
	float scale;
	int octaves;
	MyAlgebra::vec3 color0, color1;
	SolidNoise solid_noise;
};
