#pragma once

#include <vector>
#include "../my_algebra.h"

#include "GeometricObject.h"

class UniformGrid
{
public:
	UniformGrid()
		: m_pCells(0)
	{
	}

	~UniformGrid()
	{
		if (m_pCells) delete [] m_pCells;
	}

	void clear()
	{
		if (m_pCells) delete [] m_pCells;
		m_pCells = 0;
	}

	float getXCellSize() const { return m_DeltaX; }
	float getYCellSize() const { return m_DeltaY; }
	float getZCellSize() const { return m_DeltaZ; }

	int getNumXCells() const { return m_NumXCells; }
	int getNumYCells() const { return m_NumYCells; }
	int getNumZCells() const { return m_NumZCells; }

	bool init( const MyAlgebra::vec3 &bboxMin, const MyAlgebra::vec3 &bboxMax, int nXCells, int nYCells, int nZCells );

	void addObject( const GeometricObject *o );

	inline std::vector<GeometricObject *> &getCell(int xi, int yi, int zi) const
	{
		return m_pCells[xi + m_NumXCells*(yi+m_NumYCells*zi)];
	}

	inline std::vector<GeometricObject *> &getCell(int xi, int yi, int zi)
	{
		return m_pCells[xi + m_NumXCells*(yi+m_NumYCells*zi)];
	}

	inline std::vector<GeometricObject *> &getCell(int idx) const
	{
		return m_pCells[idx];
	}

	inline std::vector<GeometricObject *> &getCell(int idx)
	{
		return m_pCells[idx];
	}

	inline void getIndex(const MyAlgebra::vec3 &p, int &xi, int &yi, int &zi)
	{
		const MyAlgebra::vec3 P = p - m_BboxMin;
		xi = std::max(0, std::min((int)floorf(P.x/m_DeltaX), m_NumXCells-1));
		yi = std::max(0, std::min((int)floorf(P.y/m_DeltaY), m_NumYCells-1));
		zi = std::max(0, std::min((int)floorf(P.z/m_DeltaZ), m_NumZCells-1));
	}

	MyAlgebra::vec3 getBboxMin() const { return m_BboxMin; }
	MyAlgebra::vec3 getBboxMax() const { return m_BboxMax; }

	bool isInside(const MyAlgebra::vec3 &p) const
	{
		return (m_BboxMin.x <= p.x && p.x <= m_BboxMax.x &&
				m_BboxMin.y <= p.y && p.y <= m_BboxMax.y &&
				m_BboxMin.z <= p.z && p.z <= m_BboxMax.z );
	}

	void drawWireframe() const;

	void drawCellWireframe(int xi, int yi, int zi) const;
	void drawCellWireframe(int idx) const;

private:
	int m_NumXCells, m_NumYCells, m_NumZCells;
	MyAlgebra::vec3 m_BboxMin, m_BboxMax;
	float m_DeltaX, m_DeltaY, m_DeltaZ;

	std::vector<GeometricObject *> *m_pCells;
};

