#pragma once

#include "Ray.h"
#include "UniformGrid.h"

class UniformGridTraverser 
{
public:
	UniformGridTraverser(const Ray &ray, const UniformGrid &grid);

	~UniformGridTraverser()
	{
	}

	std::vector<GeometricObject *> &getCurrentCell() { return m_pGrid->getCell(m_CurrentX, m_CurrentY, m_CurrentZ); }

	bool isInsideGrid() const
	{
		return ( 0 <= m_CurrentX && m_CurrentX < m_pGrid->getNumXCells() &&
			     0 <= m_CurrentY && m_CurrentY < m_pGrid->getNumYCells() &&
				 0 <= m_CurrentZ && m_CurrentZ < m_pGrid->getNumZCells() );
	}

	inline float getNextRayParameter() const { return std::min(m_NextTX, std::min(m_NextTY, m_NextTZ)); }

	void advanceToNext();

	inline int getCurrentCellIndex() const
	{
		return m_CurrentX + m_pGrid->getNumXCells() * (m_CurrentY + m_pGrid->getNumYCells() * m_CurrentZ);
	}

	inline int getCurrentX() const { return m_CurrentX; }
	inline int getCurrentY() const { return m_CurrentY; }
	inline int getCurrentZ() const { return m_CurrentZ; }

	float getCurrentCellTMin() const { return m_CurrentTMin; }
	float getCurrentCellTMax() const { return std::min(m_NextTX, std::min(m_NextTY, m_NextTZ)); }

private:
	UniformGrid *m_pGrid;

	int m_CurrentX, m_CurrentY, m_CurrentZ;
	int m_IncrementX, m_IncrementY, m_IncrementZ;

	float m_CurrentTMin;
	float m_NextTX, m_NextTY, m_NextTZ;

	float m_DeltaTX, m_DeltaTY, m_DeltaTZ;

	void calculateNextIntersectionParameters(const MyAlgebra::vec3 &pIntersect, const MyAlgebra::vec3 &dir, float tStart, float &nextTX, float &nextTY, float &nextTZ) const;

	bool calculateRayBoxIntersection(float &tIntersect, const Ray &ray) const;

};

