//#undef UNICODE
#include <cstdlib>
#include <GL/glut.h>
#include "WinUtil.h"
#include <windows.h>
#include <direct.h>
#include <iostream>

// See http://wisdom.sakura.ne.jp/system/winapi/common/ for details (common controls).

using namespace std;

wchar_t *getWcharString(const char *cStr)
{
	size_t strLength = strlen( cStr );
	wchar_t *wc = new wchar_t[ strLength+1 ];

	strLength = mbstowcs( wc, cStr, strLength+1 );

	if ( strLength == -1 )
	{
		cerr << __FUNCTION__ << ": cannot convert wchar" << endl;
		return NULL;
	}

	return wc;
}

std::string
WinUtil::OpenFileDialog(const char *ext, const char *filter, bool fullpath)
{
	const GLuint window_id = glutGetWindow();

	// JgfBNg̕ۑ
	const int current_dir_buf_size = 512;
	char current_dir_buf[current_dir_buf_size];
	::_getcwd(current_dir_buf, current_dir_buf_size);

	const int max_file_path_length = 512;
	const int max_file_basename_length = 256;
	static char file_path[max_file_path_length];
	static char file_basename[max_file_basename_length];

	OPENFILENAME ofn;
	HWND hWnd = (HWND)WindowFromDC(wglGetCurrentDC());

	memset(&ofn, 0, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hWnd;
	ofn.lpstrDefExt = getWcharString( ext );
	ofn.lpstrFilter = getWcharString( filter );
	ofn.lpstrFilter = L"All Files(*.*)\0*.*\0\0";
	ofn.lpstrFile = getWcharString( file_path );
	ofn.nMaxFile = max_file_path_length;
	ofn.Flags = OFN_FILEMUSTEXIST;
	ofn.nMaxFileTitle = max_file_basename_length;
	ofn.lpstrFileTitle = getWcharString( file_basename );
	ofn.lpstrTitle = L"Select input file";

	std::string filename;

	if (GetOpenFileName(&ofn))
	{
		filename = (fullpath) ? file_path : file_basename;
		InvalidateRect(hWnd, NULL, TRUE);
	}
	else
	{
		filename = "";
	}

	glutSetWindow( window_id );

	// JgfBNg̕
	::_chdir(current_dir_buf);

	return filename;
}

std::string
WinUtil::SaveFileDialog(const char *ext, const char *filter, bool fullpath)
{
	const GLuint window_id = glutGetWindow();

	// JgfBNg̕ۑ
	const int current_dir_buf_size = 512;
	char current_dir_buf[current_dir_buf_size];
	::_getcwd(current_dir_buf, current_dir_buf_size);

	const int max_file_path_length = 512;
	const int max_file_basename_length = 256;
	static char file_path[max_file_path_length];
	static char file_basename[max_file_basename_length];

	OPENFILENAME ofn;
	HWND hWnd = (HWND)WindowFromDC(wglGetCurrentDC());

	memset(&ofn, 0, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hWnd;
	ofn.lpstrDefExt = getWcharString( ext );
	ofn.lpstrFilter = getWcharString( filter );
	ofn.lpstrFilter = L"All Files(*.*)\0*.*\0\0";
	ofn.lpstrFile = getWcharString( file_path );
	ofn.nMaxFile = max_file_path_length;
	ofn.Flags = OFN_FILEMUSTEXIST | OFN_OVERWRITEPROMPT;
	ofn.nMaxFileTitle = max_file_basename_length;
	ofn.lpstrFileTitle = getWcharString( file_basename );
	ofn.lpstrTitle = L"Save file as";

	std::string filename;

	if (GetSaveFileName(&ofn)) {
		filename = (fullpath) ? file_path : file_basename;
		//InvalidateRect(hWnd, NULL, TRUE);
	} else {
		filename = "";
	}

	glutSetWindow( window_id );

	// JgfBNg̕
	::_chdir(current_dir_buf);

	return filename;
}

void
WinUtil::ChooseColorDialog(unsigned char *color)
{
	const GLuint window_id = glutGetWindow();

	CHOOSECOLOR     cc;             // common dialog box structure
	// array of custom colors
	static COLORREF acrCustClr[16] = { // basic 16 colors
		0x0000FF, /* red */		0x00FFFF,	/* yellow */
		0x00FF00, /* lime */	0xFFFF00,	/* aqua */
		0xFF0000, /* blue */	0xFF00FF,	/* fuchsia */	
		0x000080, /* maroon */	0x008080,	/* olive */
		0x008000, /* green */	0x008000,	/* teal */
		0x800000, /* navy */	0x800080,	/* purple */
		0x000000, /* black */	0x808080,	/* gray */
		0xC0C0C0, /* silver */	0xFFFFFF	/* white */
	};

	// Initialize CHOOSECOLOR 
	ZeroMemory(&cc, sizeof(cc));
	cc.lStructSize = sizeof(cc);
	cc.hwndOwner = (HWND)WindowFromDC(wglGetCurrentDC());
	cc.lpCustColors = (LPDWORD) acrCustClr;
	cc.rgbResult = ((int)color[2] << 16) + ((int)color[1] << 8) + (int)color[0];
	cc.Flags = CC_FULLOPEN | CC_RGBINIT;

	if (ChooseColor(&cc) == TRUE) 
	{
		color[2] = (cc.rgbResult >> 16) & 255;
		color[1] = (cc.rgbResult >>  8) & 255;
		color[0] = (cc.rgbResult      ) & 255;
	}

	glutSetWindow( window_id );
}
