#pragma once

#include "Texture.h"
#include "SolidNoise.h"

class WoodenTexture : public Texture
{
private:
	WoodenTexture()
		: scale(50.f), octaves(2),
		color0(0.2f, 0.1f, 0.1f), color1(0.8f, 0.6f, 0.4f)
	{
		float stripes_per_unit = 0.15f;
		freq = (float)M_PI * stripes_per_unit;
	}

public:
	static WoodenTexture *CreateTexture()
	{
		WoodenTexture *t = new WoodenTexture();
		s_TextureCache.push_back( t );
		return t;
	}

	MyAlgebra::vec3 getColor(const MyAlgebra::vec3 &uvw) const
	{
		const float temp = scale * fabs( solid_noise.turbulence(freq*uvw, octaves) );
		const int tInt = (int)temp;
		const float t = temp - tInt;

		return t * color0 + (1.f-t) * color1;
	}

	float getScale() const { return scale; }
	void setScale(float s) { scale = s; }
	void setStripesPerUnit(float p) { freq = (float)M_PI * p; }

	void setColor0(const MyAlgebra::vec3 &c) { color0.set(c); }
	void setColor1(const MyAlgebra::vec3 &c) { color1.set(c); }
	void setColor0(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color0.set(r, g, b); }
	void setColor1(MyAlgebra::Real r, MyAlgebra::Real g, MyAlgebra::Real b) { color1.set(r, g, b); }

	MyAlgebra::vec3 getColor0() const { return color0; }
	MyAlgebra::vec3 getColor1() const { return color1; }

	Texture_Type getTextureType() const { return Wooden_Texture; }
	bool isSolidTexture() const { return true; }

private:
	float freq, scale;
	int octaves;
	MyAlgebra::vec3 color0, color1;
	SolidNoise solid_noise;
};
