#include "WorldViewWindow.h"
#include "Scene.h"
#include "ImageRect.h"
#include "RayTracer.h"
#include "EnvironmentMap.h"

using namespace std;
using namespace MyAlgebra;

extern RayTracer *g_pRayTracer;
extern ImageRect<vec3> g_ScreenBuffer;
extern Camera g_Camera;
extern Scene g_Scene;

extern int g_KeepRayTracing;

int WorldViewWindow::s_ToggleWireframe = 0;
int WorldViewWindow::s_DrawObjects = 1;
int WorldViewWindow::s_DrawViewFrustum = 1;
int WorldViewWindow::s_DrawWorldAxes = 1;

extern vec3 g_DebugHitPoint;

void WorldViewWindow::display()
{
	glClearColor(1,1,1,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushAttrib(GL_ENABLE_BIT);

	glEnable(GL_DEPTH_TEST);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	m_Camera.setGLPerspective();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	m_Camera.setGLModelview();

	if (s_DrawObjects)
	{
		if (s_ToggleWireframe)
		{
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			glLineWidth( 1.f );
		}

		for (int i=0; i<g_Scene.getNumObjects(); i++)
		{
			glColor3fv( g_Scene.getPseudoColor(i) );
			g_Scene.getObject(i)->drawGL();
		}

		if (s_ToggleWireframe)
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}

	// axes

	if (s_DrawWorldAxes)
	{
		const float scale = 1.f;

		glLineWidth(3.f);

		glBegin(GL_LINES);
		glColor3f(1,0,0); glVertex3f(0,0,0); glVertex3f(scale,0,0);
		glColor3f(0,1,0); glVertex3f(0,0,0); glVertex3f(0,scale,0);
		glColor3f(0,0,1); glVertex3f(0,0,0); glVertex3f(0,0,scale);
		glEnd();
	}

	// camera view frustum

	if (s_DrawViewFrustum)
		displayViewFrustum();

	// debug hit point

	glColor3f(0,1,0);
	glLineWidth(5.f);

	glBegin(GL_LINES);
	glVertex3fv( g_Camera.getEyePosition() );
	glVertex3fv( g_DebugHitPoint );
	glEnd();

	glPopAttrib();

	glutSwapBuffers();
}

void WorldViewWindow::reshape(int w, int h)
{
	if (h <= 1) return;

	m_WindowWidth = w;
	m_WindowHeight = h;

	glViewport(0, 0, w, h);
	m_Camera.setWindowSize(w,h);
}

void WorldViewWindow::keyboard(unsigned char key, int x, int y)
{
	switch(key)
	{
	case 'c':
		{
			vec3 eye, center, up;
			m_Camera.getLookAt(eye, center, up);

			cout << __FUNCTION__ << ": camera parameters:" << endl
				<< "  eye=" << eye << endl
				<< "  center=" << center << endl
				<< "  up=" << up << endl;
		}
		break;
	default:
		break;
	}
}

void WorldViewWindow::mouse(int button, int state, int x, int _y)
{
	if ( ! m_Camera.isGrabbed() && state == GLUT_DOWN)
	{
		if (button == GLUT_LEFT_BUTTON)
			m_Camera.grab(Camera::ROTATE_ACTION, x, _y);
		else if (button == GLUT_MIDDLE_BUTTON)
			m_Camera.grab(Camera::ZOOM_ACTION, x, _y);
		else if (button == GLUT_RIGHT_BUTTON)
			m_Camera.grab(Camera::PAN_ACTION, x, _y);
	}
	else if (state == GLUT_UP)
	{
		if (m_Camera.isGrabbed()) m_Camera.release();
		display();
	}
}

void WorldViewWindow::motion(int x, int _y)
{
	if (m_Camera.isGrabbed())
	{
		m_Camera.roll(x,_y);
		display();
	}
}

void WorldViewWindow::displayViewFrustum() const
{
	const vec3 eye = g_Camera.getEyePosition();

	const float w = g_Camera.getWindowWidth();
	const float h = g_Camera.getWindowHeight();
	const float d = g_Camera.calcScreenDistance();

	const float dFixed = 2.f;

	const float wHalf = 0.5f * w * (dFixed/d);
	const float hHalf = 0.5f * h * (dFixed/d);

	vec3 xAxis, yAxis, zAxis;
	g_Camera.getModelviewCoordinates(xAxis, yAxis, zAxis);

	const vec3 frame0 = eye - wHalf*xAxis - hHalf*yAxis - dFixed*zAxis;
	const vec3 frame1 = eye + wHalf*xAxis - hHalf*yAxis - dFixed*zAxis;
	const vec3 frame2 = eye + wHalf*xAxis + hHalf*yAxis - dFixed*zAxis;
	const vec3 frame3 = eye - wHalf*xAxis + hHalf*yAxis - dFixed*zAxis;

	glColor3f(0,0,0);
	glLineWidth(3.f);

	glEnable(GL_LINE_STIPPLE);
	glLineStipple(1, 0x0f0f);

	glBegin(GL_LINES);
	glVertex3fv(eye);	glVertex3fv(frame0);	
	glVertex3fv(eye);	glVertex3fv(frame1);	
	glVertex3fv(eye);	glVertex3fv(frame2);	
	glVertex3fv(eye);	glVertex3fv(frame3);	
	glEnd();

	glDisable(GL_LINE_STIPPLE);

	glBegin(GL_LINE_LOOP);
	glVertex3fv(frame0);
	glVertex3fv(frame1);
	glVertex3fv(frame2);
	glVertex3fv(frame3);
	glEnd();
}

