#include "RecursiveRayTracer.h"

#include "Camera.h"
#include "Scene.h"
#include "SceneLoader.h"

#include "ConfigLoader.h"

#include "Sphere.h"
#include "Plane.h"
#include "AxisAlignedBox.h"
#include "Triangle.h"
#include "TriangleMesh.h"
#include "MetaballSurface.h"
#include "GeometryInstance.h"

#include "TexturedSphere.h"
#include "SolidTexturedSphere.h"
#include "TexturedPlane.h"

#include "PointLight.h"
#include "DirectionalLight.h"
#include "EnvironmentLight.h"

#include "PseudoNormalColorMaterial.h"
#include "DiffuseMaterial.h"
#include "PhongMaterial.h"
#include "TexturedMaterial.h"
#include "PerfectSpecularMaterial.h"
#include "SpecularRefractionMaterial.h"

#include "CheckeredTexture.h"
#include "SolidCheckeredTexture.h"
#include "NoiseTexture.h"
#include "MarbleTexture.h"
#include "WoodenTexture.h"
#include "TurbulenceTexture.h"
#include "DerivTurbulenceTexture.h"
#include "ImageTexture.h"

#include "BoxFilter.h"

#include "SingleSampler.h"
#include "RandomSampler.h"
#include "Regular2x2Sampler.h"
#include "Regular3x3Sampler.h"
#include "RegularNxNSampler.h"

#include "EnvironmentMap.h"
#include "EnvironmentLightGenerator.h"

#include "MeshLoader.h"

#include "ImageRect.h"
#include "stopwatch.h"
#include "OGLDisplayDriver.h"

#include "DebugIntersectionLog.h"

#include "MiscSettings.h"

using namespace std;
using namespace MyAlgebra;

// global variables

RayTracer *g_pRayTracer = 0;

ImageRect<vec3> g_ScreenBuffer(400, 300);
Camera g_Camera;
Scene g_Scene;

MiscSettings g_MiscSettings;

bool g_IsBatchMode = false;
string g_CommandlineScreenshotFilename = "";

// for debugging

DebugIntersectionLog debugLog;
int g_CurrentXi, g_CurrentYi;

inline float frand() { return 2.f*rand()/(float)RAND_MAX; }

void parseCommandlineArgs(int argc, char **argv)
{
	if (argc == 1) return;

	for (int i=1; i<argc; i++)
	{
		string token(argv[i]);

		//cerr << __FUNCTION__ << ": token[" << i << "] = " << token << endl;

		if ( "-b" == token)
		{
			g_IsBatchMode = true;
			cerr << __FUNCTION__ << ": entering batch mode" << endl;
		}
		else if ( "-f" == token)
		{
			if (++i < argc)
			{
				g_MiscSettings.m_SceneFilename = argv[i];
				cerr << __FUNCTION__ << ": setting scene filename: " << g_MiscSettings.m_SceneFilename << endl;
			}
			else
			{
				cerr << __FUNCTION__ << ": missing scene filename" << endl;
				break;
			}
		}
		else if ( "-o" == token)
		{
			if (++i < argc)
			{
				g_CommandlineScreenshotFilename = argv[i];
				cerr << __FUNCTION__ << ": setting screenshot filename: " << g_CommandlineScreenshotFilename << endl;
			}
			else
			{
				cerr << __FUNCTION__ << ": missing screenshot filename" << endl;
				break;
			}
		}
		else
		{
			cerr << __FUNCTION__ << ": unknown token: " << token << endl;
			break;
		}
	}
}

int main(int argc, char **argv)
{
	// OpenGL display driver
	OGLDisplayDriver::Init(argc, argv);
	
	g_MiscSettings.init();

	{
		data_path filePath;
		filePath.path.push_back("..");
		filePath.path.push_back("../..");
		filePath.path.push_back("../../..");

		string filename = "earth_sphere.jpg";

		FILE *fp = fopen(filePath.get_file(filename).c_str(), "rb");

		if ( fp == NULL )
		{
			cerr << "test cannot open image" << endl;
		}
		else
		{
			cerr << "test open image success" << endl;
			fclose(fp);
		}
	}

	parseCommandlineArgs(argc, argv);

	g_pRayTracer = new RecursiveRayTracer();

	{
		ConfigLoader loader;

		loader.load( g_MiscSettings.m_SceneFilename.c_str(), 1==g_MiscSettings.m_SceneVerboseLoad );

		if ( "" != g_CommandlineScreenshotFilename)
			g_MiscSettings.m_ScreenshotFilename = g_CommandlineScreenshotFilename;
	}

	g_pRayTracer->registerImageFilter( new BoxFilter() );
	g_pRayTracer->registerImageSampler( g_MiscSettings.m_pSampler );

	if ( g_IsBatchMode )
	{
		cerr << "ExperimentalRayTracer: entering batch mode" << endl;
		g_MiscSettings.m_DumpComputationTime = 1;

		g_pRayTracer->renderScene( g_ScreenBuffer );

		ImageIO io;
		io.save( g_MiscSettings.m_ScreenshotFilename.c_str(), g_ScreenBuffer );

		exit(0);
	}

	OGLDisplayDriver::MainLoop();

	return 0;
}

